# mypy: ignore-errors
"""
===========================
Framework Utility Functions
===========================

Collection of utility functions shared by the ``vivarium`` framework.

"""
import functools
from bdb import BdbQuit
from collections.abc import Callable, Sequence
from importlib import import_module
from typing import Any, Literal, TypeVar

import numpy as np
from loguru import logger

from vivarium.types import NumberLike, NumericArray, Timedelta

TimeValue = TypeVar("T", bound=NumberLike)


def from_yearly(value: TimeValue, time_step: Timedelta) -> TimeValue:
    return value * (time_step.total_seconds() / (60 * 60 * 24 * 365.0))


def to_yearly(value: TimeValue, time_step: Timedelta) -> TimeValue:
    return value / (time_step.total_seconds() / (60 * 60 * 24 * 365.0))


def rate_to_probability(
    rate: Sequence[float] | NumberLike,
    time_scaling_factor: float | int = 1.0,
    rate_conversion_type: Literal["linear", "exponential"] = "linear",
) -> NumericArray:
    """Converts a rate to a probability.

    Parameters
    ----------
    rate
        The rate to convert to a probability.
    time_scaling_factor
        The time factor in to scale the rates by. This is usually the time step.
    rate_conversion_type
        The type of conversion to use. Default is "linear" for a simple multiplcation
        of rate and time_scaling_factor. The other option is "exponential" which should be
        used for continuous time event driven models.

    Returns
    -------
        An array of floats representing the probability of the converted rates
    """
    if rate_conversion_type not in ["linear", "exponential"]:
        raise ValueError(
            f"Rate conversion type {rate_conversion_type} is not implemented. "
            "Allowable types are 'linear' or 'exponential'."
        )
    if rate_conversion_type == "linear":
        # NOTE: The default behavior for randomness streams is to use a rate that is already
        # scaled to the time step which is why the default time scaling factor is 1.0.
        probability = np.array(rate * time_scaling_factor)

        # Clip to 1.0 if the probability is greater than 1.0.
        exceeds_one = probability > 1.0
        if exceeds_one.any():
            probability[exceeds_one] = 1.0
            logger.warning(
                "The rate to probability conversion resulted in a probability greater than 1.0. "
                "The probability has been clipped to 1.0 and indicates the rate is too high. "
            )
    else:
        # encountered underflow from rate > 30k
        # for rates greater than 250, exp(-rate) evaluates to 1e-109
        # beware machine-specific floating point issues
        rate = np.array(rate)
        rate[rate > 250] = 250.0
        probability: NumericArray = 1 - np.exp(-rate * time_scaling_factor)

    return probability


def probability_to_rate(
    probability: Sequence[float] | NumberLike,
    time_scaling_factor: float | int = 1.0,
    rate_conversion_type: Literal["linear", "exponential"] = "linear",
) -> NumericArray:
    """Function to convert a probability to a rate.

    Parameters
    ----------
    probability
        The probability to convert to a rate.
    time_scaling_factor
        The time factor in to scale the probability by. This is usually the time step.
    rate_conversion_type
        The type of conversion to use. Default is "linear" for a simple multiplcation
        of rate and time_scaling_factor. The other option is "exponential" which should be
        used for continuous time event driven models.

    Returns
    -------
        An array of floats representing the rate of the converted probabilities
    """
    # NOTE: The default behavior for randomness streams is to use a rate that is already
    # scaled to the time step which is why the default time scaling factor is 1.0.
    if rate_conversion_type not in ["linear", "exponential"]:
        raise ValueError(
            f"Rate conversion type {rate_conversion_type} is not implemented. "
            "Allowable types are 'linear' or 'exponential'."
        )
    if rate_conversion_type == "linear":
        rate = np.array(probability / time_scaling_factor)
    else:
        probability = np.array(probability)
        rate: NumericArray = -np.log(1 - probability)
    return rate


def collapse_nested_dict(
    d: dict[str, Any], prefix: str | None = None
) -> list[tuple[str, Any]]:
    results = []
    for k, v in d.items():
        cur_prefix = prefix + "." + k if prefix else k
        if isinstance(v, dict):
            results.extend(collapse_nested_dict(v, prefix=cur_prefix))
        else:
            results.append((cur_prefix, v))
    return results


def import_by_path(path: str) -> Callable[..., Any]:
    """Import a class or function given its absolute path.

    Parameters
    ----------
    path
        Path to object to import

    Returns
    -------
        The imported class or function
    """

    module_path, _, class_name = path.rpartition(".")
    callable_attr: Callable[..., Any] = getattr(import_module(module_path), class_name)
    return callable_attr


def handle_exceptions(
    func: Callable[..., Any], logger: Any, with_debugger: bool
) -> Callable[..., Any]:
    """Drops a user into an interactive debugger if func raises an error."""

    @functools.wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except (BdbQuit, KeyboardInterrupt):
            raise
        except Exception as e:
            logger.exception("Uncaught exception {}".format(e))
            if with_debugger:
                import pdb
                import traceback

                traceback.print_exc()
                pdb.post_mortem()
            else:
                raise

    return wrapped



from __future__ import annotations

import re
from collections.abc import Callable
from types import MethodType

import pandas as pd
import pytest
from layered_config_tree.main import LayeredConfigTree
from loguru import logger
from pytest_mock import MockerFixture

from tests.framework.results.helpers import BASE_POPULATION, FAMILIARS
from tests.framework.results.helpers import HOUSE_CATEGORIES as HOUSES
from tests.framework.results.helpers import mock_get_value
from vivarium.framework.event import Event
from vivarium.framework.lifecycle import lifecycle_states
from vivarium.framework.results import ResultsInterface, ResultsManager
from vivarium.framework.results.observation import (
    ConcatenatingObservation,
    StratifiedObservation,
)


def _silly_aggregator(_: pd.DataFrame) -> float:
    return 1.0


####################################
# Test stratification registration #
####################################


def test_register_stratification(mocker: MockerFixture) -> None:
    def _silly_mapper(some_series: pd.Series[str]) -> str:
        # NOTE: it does not actually matter what this mapper returns for this test
        return "this was pointless"

    builder = mocker.Mock()
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr = ResultsManager()
    mgr.setup(builder)
    interface = ResultsInterface(mgr)

    # Check pre-registration stratifications and manager required columns/values
    assert len(mgr._results_context.stratifications) == 0

    interface.register_stratification(
        name="some-name",
        categories=["some-category", "some-other-category", "some-unwanted-category"],
        excluded_categories=["some-unwanted-category"],
        mapper=_silly_mapper,
        is_vectorized=False,
        requires_columns=["some-column", "some-other-column"],
        requires_values=["some-value", "some-other-value"],
    )

    # Check stratification registration
    stratifications = mgr._results_context.stratifications
    assert len(stratifications) == 1

    stratification = stratifications["some-name"]
    pipeline_names = [pipeline.name for pipeline in stratification.requires_values]

    assert stratification.name == "some-name"
    assert stratification.requires_columns == ["some-column", "some-other-column"]
    assert pipeline_names == ["some-value", "some-other-value"]
    assert stratification.categories == ["some-category", "some-other-category"]
    assert stratification.excluded_categories == ["some-unwanted-category"]
    assert stratification.mapper == _silly_mapper
    assert stratification.is_vectorized is False


@pytest.mark.parametrize(
    "target, target_type", [("some-column", "column"), ("some-value", "value")]
)
def test_register_binned_stratification(
    target: str, target_type: str, mocker: MockerFixture
) -> None:

    mgr = ResultsManager()
    mgr.logger = logger
    builder = mocker.Mock()
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    # mgr._results_context.setup(builder)

    # Check pre-registration stratifications and manager required columns/values
    assert len(mgr._results_context.stratifications) == 0

    mgr.register_binned_stratification(
        target=target,
        binned_column="new-binned-column",
        bin_edges=[1, 2, 3],
        labels=["1_to_2", "2_to_3"],
        excluded_categories=["2_to_3"],
        target_type=target_type,
        some_kwarg="some-kwarg",
        some_other_kwarg="some-other-kwarg",
    )

    # Check stratification registration
    stratifications = mgr._results_context.stratifications
    assert len(stratifications) == 1
    expected_column_names = [target] if target_type == "column" else []
    expected_value_names = [target] if target_type == "value" else []

    stratification = stratifications["new-binned-column"]
    assert stratification.name == "new-binned-column"
    assert stratification.requires_columns == expected_column_names
    assert [value.name for value in stratification.requires_values] == expected_value_names
    assert stratification.categories == ["1_to_2"]
    assert stratification.excluded_categories == ["2_to_3"]
    # Cannot access the mapper because it's in local scope, so check __repr__
    assert "function ResultsManager.register_binned_stratification.<locals>._bin_data" in str(
        stratification.mapper
    )
    assert stratification.is_vectorized is True


#################################
# Test observation registration #
#################################


@pytest.mark.parametrize(
    ("obs_type", "missing_args"),
    [
        ("StratifiedObservation", ["results_updater"]),
        ("UnstratifiedObservation", ["results_gatherer", "results_updater"]),
    ],
)
def test_register_observation_raises(
    obs_type: str, missing_args: list[str], mocker: MockerFixture
) -> None:
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr = ResultsManager()
    mgr.setup(builder)
    interface = ResultsInterface(mgr)
    match = re.escape(
        f"Observation 'some-name' is missing required callable(s): {missing_args}",
    )
    with pytest.raises(ValueError, match=match):
        if obs_type == "StratifiedObservation":
            interface.register_stratified_observation(name="some-name")
        if obs_type == "UnstratifiedObservation":
            interface.register_unstratified_observation(name="some-name")


def test_register_stratified_observation(mocker: MockerFixture) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = ["default-stratification", "exclude-this"]
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    for strat in [
        "default-stratification",
        "some-stratification",
        "some-other-stratification",
        "exclude-this",
    ]:
        interface.register_stratification(
            name=strat,
            categories=["a", "b", "c"],
            excluded_categories=[],
            is_vectorized=True,
            requires_columns=[strat],
        )

    assert len(interface._manager._results_context.grouped_observations) == 0

    interface.register_stratified_observation(
        name="some-name",
        pop_filter="some-filter",
        when="some-when",
        requires_columns=["some-column", "some-other-column"],
        requires_values=["some-value", "some-other-value"],
        results_updater=lambda _, __: pd.DataFrame(),
        additional_stratifications=["some-stratification", "some-other-stratification"],
        excluded_stratifications=["exclude-this"],
    )

    mgr.on_post_setup(mocker.Mock())

    observations_dict = interface._manager._results_context.observations
    assert len(observations_dict) == 1
    assert "some-name" in observations_dict

    grouped_observations = interface._manager._results_context.grouped_observations
    assert len(grouped_observations) == 1
    filter = list(grouped_observations["some-when"].keys())[0]
    stratifications = list(grouped_observations["some-when"][filter])[0]
    observations = grouped_observations["some-when"][filter][stratifications]
    assert filter == "some-filter"
    assert isinstance(stratifications, tuple)  # for mypy in following set(stratifications)
    assert set(stratifications) == {
        "default-stratification",
        "some-stratification",
        "some-other-stratification",
    }
    assert len(observations) == 1

    for observation in [observations_dict["some-name"], observations[0]]:
        assert observation.name == "some-name"
        assert observation.pop_filter == "some-filter"
        assert observation.when == "some-when"
        assert observation.results_gatherer is not None
        assert observation.results_updater is not None
        assert observation.results_formatter is not None
        assert observation.stratifications is not None
        assert {strat.name for strat in observation.stratifications} == set(stratifications)
        assert isinstance(observation, StratifiedObservation)
        assert observation.aggregator is not None
        assert observation.aggregator_sources is None


def test_register_unstratified_observation(mocker: MockerFixture) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    assert len(interface._manager._results_context.grouped_observations) == 0
    interface.register_unstratified_observation(
        name="some-name",
        pop_filter="some-filter",
        when="some-when",
        requires_columns=["some-column", "some-other-column"],
        requires_values=["some-value", "some-other-value"],
        results_gatherer=lambda _: pd.DataFrame(),
        results_updater=lambda _, __: pd.DataFrame(),
    )
    grouped_observations = interface._manager._results_context.grouped_observations
    assert len(grouped_observations) == 1
    filter = list(grouped_observations["some-when"].keys())[0]
    stratifications = list(grouped_observations["some-when"][filter])[0]
    observations = grouped_observations["some-when"][filter][stratifications]
    assert filter == "some-filter"
    assert stratifications is None
    assert len(observations) == 1
    obs = observations[0]
    assert obs.name == "some-name"
    assert obs.pop_filter == "some-filter"
    assert obs.when == "some-when"
    assert obs.results_gatherer is not None
    assert obs.results_updater is not None
    assert obs.results_formatter is not None


@pytest.mark.parametrize(
    (
        "name, pop_filter, aggregator_columns, aggregator, requires_columns, requires_values,"
        " additional_stratifications, excluded_stratifications, when"
    ),
    [
        (
            "living_person_time",
            'alive == "alive" and undead == False',
            [],
            _silly_aggregator,
            [],
            [],
            [],
            [],
            lifecycle_states.TIME_STEP_CLEANUP,
        ),
        (
            "undead_person_time",
            "undead == True",
            [],
            _silly_aggregator,
            [],
            [],
            [],
            [],
            lifecycle_states.TIME_STEP_PREPARE,
        ),
        (
            "undead_person_time",
            "undead == True",
            [],
            _silly_aggregator,
            [],
            ["fake_pipeline", "another_fake_pipeline"],
            [],
            [],
            lifecycle_states.TIME_STEP_PREPARE,
        ),
    ],
    ids=["valid_on_collect_metrics", "valid_on_time_step__prepare", "valid_pipelines"],
)
def test_register_adding_observation(
    mocker: MockerFixture,
    name: str,
    pop_filter: str,
    aggregator_columns: list[str],
    aggregator: Callable[[pd.DataFrame], int | float | pd.Series[int | float]],
    requires_columns: list[str],
    requires_values: list[str],
    additional_stratifications: list[str],
    excluded_stratifications: list[str],
    when: str,
) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    assert len(interface._manager._results_context.grouped_observations) == 0
    interface.register_adding_observation(
        name=name,
        pop_filter=pop_filter,
        when=when,
        additional_stratifications=additional_stratifications,
        excluded_stratifications=excluded_stratifications,
        aggregator_sources=aggregator_columns,
        aggregator=aggregator,
        requires_columns=requires_columns,
        requires_values=requires_values,
    )
    assert len(interface._manager._results_context.grouped_observations) == 1


def test_register_multiple_adding_observations(mocker: MockerFixture) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr.setup(builder)

    assert len(interface._manager._results_context.grouped_observations) == 0
    interface.register_adding_observation(
        name="living_person_time",
        when=lifecycle_states.TIME_STEP_CLEANUP,
        aggregator=_silly_aggregator,
    )
    # Test observation gets added
    assert len(interface._manager._results_context.grouped_observations) == 1
    assert (
        interface._manager._results_context.grouped_observations[
            lifecycle_states.TIME_STEP_CLEANUP
        ]["tracked==True"][()][0].name
        == "living_person_time"
    )

    interface.register_adding_observation(
        name="undead_person_time",
        pop_filter="undead==True",
        when=lifecycle_states.TIME_STEP_PREPARE,
        aggregator=_silly_aggregator,
    )
    # Test new observation gets added
    assert len(interface._manager._results_context.grouped_observations) == 2
    assert (
        interface._manager._results_context.grouped_observations[
            lifecycle_states.TIME_STEP_CLEANUP
        ]["tracked==True"][()][0].name
        == "living_person_time"
    )
    assert (
        interface._manager._results_context.grouped_observations[
            lifecycle_states.TIME_STEP_PREPARE
        ]["undead==True"][()][0].name
        == "undead_person_time"
    )


@pytest.mark.parametrize("resource_type", ["value", "column"])
def test_unhashable_pipeline(mocker: MockerFixture, resource_type: str) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    mgr.setup(builder)

    assert len(interface._manager._results_context.grouped_observations) == 0
    with pytest.raises(TypeError, match=f"All required {resource_type}s must be strings"):
        interface.register_adding_observation(
            name="living_person_time",
            pop_filter='alive == "alive" and undead == False',
            when=lifecycle_states.TIME_STEP_CLEANUP,
            requires_columns=[["bad", "unhashable", "thing"]] if resource_type == "column" else [],  # type: ignore[list-item]
            requires_values=[["bad", "unhashable", "thing"]] if resource_type == "value" else [],  # type: ignore[list-item]
            additional_stratifications=[],
            excluded_stratifications=[],
            aggregator_sources=[],
            aggregator=_silly_aggregator,
        )


@pytest.mark.parametrize(
    "when",
    [
        lifecycle_states.TIME_STEP_PREPARE,
        lifecycle_states.TIME_STEP,
        lifecycle_states.TIME_STEP_CLEANUP,
        lifecycle_states.COLLECT_METRICS,
    ],
)
def test_register_adding_observation_when_options(when: str, mocker: MockerFixture) -> None:
    """Test the full interface lifecycle of adding an observation and simulation event."""
    mgr = ResultsManager()
    results_interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    mgr.setup(builder)
    mgr.population_view = mocker.Mock()
    mgr.population_view.subview.return_value = mgr.population_view  # type: ignore[attr-defined]
    mgr.population_view.get.return_value = BASE_POPULATION.copy()  # type: ignore[attr-defined]

    # register stratifications
    results_interface.register_stratification(
        name="house", categories=HOUSES, is_vectorized=True, requires_columns=["house"]
    )
    results_interface.register_stratification(
        name="familiar",
        categories=FAMILIARS,
        is_vectorized=True,
        requires_columns=["familiar"],
    )

    aggregator_map = {
        lifecycle_state: mocker.Mock(side_effect=lambda x: 1.0)
        for lifecycle_state in [
            lifecycle_states.TIME_STEP_PREPARE,
            lifecycle_states.TIME_STEP,
            lifecycle_states.TIME_STEP_CLEANUP,
            lifecycle_states.COLLECT_METRICS,
        ]
    }

    # Register observations to all four phases
    for phase, aggregator in aggregator_map.items():
        results_interface.register_adding_observation(
            name=f"{phase}_measure",
            when=phase,
            additional_stratifications=["house", "familiar"],
            aggregator=aggregator,
            requires_columns=["house", "familiar"],
        )

    for mock_aggregator in aggregator_map.values():
        mock_aggregator.assert_not_called()

    # Fake a timestep
    event = Event(
        name=when,
        index=pd.Index([0]),
        user_data={},
        time=0,
        step_size=1,
    )
    # Run on_post_setup to initialize the raw_results attribute with 0s and set stratifications
    mgr.on_post_setup(event)
    mgr.gather_results(event)

    for phase, aggregator in aggregator_map.items():
        if phase == when:
            aggregator.assert_called()
        else:
            aggregator.assert_not_called()


def test_register_concatenating_observation(mocker: MockerFixture) -> None:
    mgr = ResultsManager()
    interface = ResultsInterface(mgr)
    builder = mocker.Mock()
    builder.configuration.stratification.default = []
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    assert len(interface._manager._results_context.grouped_observations) == 0
    interface.register_concatenating_observation(
        name="some-name",
        pop_filter="some-filter",
        when="some-when",
        requires_columns=["some-column", "some-other-column"],
        requires_values=["some-value", "some-other-value"],
        results_formatter=lambda _, __: pd.DataFrame(),
    )
    grouped_observations = interface._manager._results_context.grouped_observations
    assert len(grouped_observations) == 1
    filter = list(grouped_observations["some-when"].keys())[0]
    stratifications = list(grouped_observations["some-when"][filter])[0]
    observations = grouped_observations["some-when"][filter][stratifications]
    assert filter == "some-filter"
    assert stratifications is None
    assert len(observations) == 1
    obs = observations[0]
    assert obs.name == "some-name"
    assert obs.pop_filter == "some-filter"
    assert obs.when == "some-when"
    assert isinstance(obs, ConcatenatingObservation)
    assert obs.included_columns == [
        "event_time",
        "some-column",
        "some-other-column",
        "some-value",
        "some-other-value",
    ]
    assert obs.results_gatherer is not None
    assert obs.results_updater is not None
    assert obs.results_formatter is not None



"""
=====================
LookupTable Interface
=====================

This module provides an interface to the :class:`LookupTableManager <vivarium.framework.lookup.manager.LookupTableManager>`.

"""


from collections.abc import Sequence

from vivarium.framework.lookup.manager import LookupTableManager
from vivarium.framework.lookup.table import LookupTable
from vivarium.manager import Interface
from vivarium.types import LookupTableData


class LookupTableInterface(Interface):
    """The lookup table management system.

    Simulations tend to require a large quantity of data to run. ``vivarium``
    provides the :class:`Lookup Table <vivarium.framework.lookup.table.LookupTable>`
    abstraction to ensure that accurate data can be retrieved when it's needed.

    For more information, see :ref:`here <lookup_concept>`.

    """

    def __init__(self, manager: LookupTableManager):
        self._manager = manager

    def build_table(
        self,
        data: LookupTableData,
        key_columns: Sequence[str] = (),
        parameter_columns: Sequence[str] = (),
        value_columns: Sequence[str] = (),
    ) -> LookupTable:
        """Construct a LookupTable from input data.

        If data is a :class:`pandas.DataFrame`, an interpolation function of
        the order specified in the simulation
        :term:`configuration <Configuration>` will be calculated for each
        permutation of the set of key_columns. The columns in parameter_columns
        will be used as parameters for the interpolation functions which will
        estimate all remaining columns in the table.

        If data is a number, time, list, or tuple, a scalar table will be
        constructed with the values in data as the values in each column of
        the table, named according to value_columns.


        Parameters
        ----------
        data
            The source data which will be used to build the resulting
            :class:`Lookup Table <vivarium.framework.lookup.table.LookupTable>`.
        key_columns
            Columns used to select between interpolation functions. These
            should be the non-continuous variables in the data. For example
            'sex' in data about a population.
        parameter_columns
            The columns which contain the parameters to the interpolation
            functions. These should be the continuous variables. For example
            'age' in data about a population.
        value_columns
            The data columns that will be in the resulting LookupTable. Columns
            to be interpolated over if interpolation or the names of the columns
            in the scalar table.

        Returns
        -------
            LookupTable
        """
        return self._manager.build_table(data, key_columns, parameter_columns, value_columns)



"""
==================
Life Cycle Manager
==================

Manager of ordering- and constraint-based contracts in the simulation

"""
from __future__ import annotations

import time
from collections import defaultdict
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from vivarium.framework.lifecycle.constraint_maker import ConstraintMaker
from vivarium.framework.lifecycle.entities import LifeCycle
from vivarium.framework.lifecycle.exceptions import InvalidTransitionError, LifeCycleError
from vivarium.framework.lifecycle.lifecycle_states import INITIALIZATION
from vivarium.manager import Manager

if TYPE_CHECKING:
    from vivarium.framework.event import Event


class LifeCycleManager(Manager):
    """Manages ordering- and constraint-based contracts in the simulation."""

    def __init__(self) -> None:
        self.lifecycle = LifeCycle()
        self._current_state = self.lifecycle.get_state(INITIALIZATION)
        self._current_state_start_time = time.time()
        self._timings: defaultdict[str, list[float]] = defaultdict(list)
        self._make_constraint = ConstraintMaker(self)

    @property
    def name(self) -> str:
        """The name of this component."""
        return "life_cycle_manager"

    @property
    def current_state(self) -> str:
        """The name of the current life cycle state."""
        return self._current_state.name

    @property
    def timings(self) -> dict[str, list[float]]:
        return self._timings

    def add_phase(self, phase_name: str, states: list[str], loop: bool = False) -> None:
        """Add a new phase to the lifecycle.

        Phases must be added in order.

        Parameters
        ----------
        phase_name
            The name of the phase to add.  Phase names must be unique.
        states
            The list of names (in order) of the states that make up the
            life cycle phase.  State names must be unique across the entire
            life cycle.
        loop
            Whether the life cycle phase states loop.

        Raises
        ------
        LifeCycleError
            If the phase or state names are non-unique.
        """
        self.lifecycle.add_phase(phase_name, states, loop)

    def set_state(self, state: str) -> None:
        """Sets the current life cycle state to the provided state.

        Parameters
        ----------
        state
            The name of the state to set.

        Raises
        ------
        LifeCycleError
            If the requested state doesn't exist in the life cycle.
        InvalidTransitionError
            If setting the provided state represents an invalid life cycle
            transition.
        """
        new_state = self.lifecycle.get_state(state)
        if self._current_state.valid_next_state(new_state):
            self._timings[self._current_state.name].append(
                time.time() - self._current_state_start_time
            )
            new_state.enter()
            self._current_state = new_state
            self._current_state_start_time = time.time()
        else:
            raise InvalidTransitionError(
                f"Invalid transition from {self.current_state} "
                f"to {new_state.name} requested."
            )

    def get_state_names(self, phase: str) -> list[str]:
        """Gets all states in the phase in their order of execution.

        Parameters
        ----------
        phase
            The name of the phase to retrieve the states for.

        Returns
        -------
            A list of state names in order of execution.
        """
        return self.lifecycle.get_state_names(phase)

    def add_handlers(self, state_name: str, handlers: list[Callable[[Event], None]]) -> None:
        """Registers a set of functions to be called during a life cycle state.

        This method does not apply any constraints, rather it is used
        to build up an execution order for introspection.

        Parameters
        ----------
        state_name
            The name of the state to register the handlers for.
        handlers
            A list of functions that will execute during the state.
        """
        s = self.lifecycle.get_state(state_name)
        s.add_handlers(handlers)

    def add_constraint(
        self,
        method: Callable[..., Any],
        allow_during: tuple[str, ...] | list[str] = (),
        restrict_during: tuple[str, ...] | list[str] = (),
    ) -> None:
        """Constrains a function to be executable only during certain states.

        Parameters
        ----------
        method
            The method to add constraints to.
        allow_during
            An optional list of life cycle states in which the provided
            method is allowed to be called.
        restrict_during
            An optional list of life cycle states in which the provided
            method is restricted from being called.

        Raises
        ------
        ValueError
            If neither ``allow_during`` nor ``restrict_during`` are provided,
            or if both are provided.
        LifeCycleError
            If states provided as arguments are not in the life cycle.
        ConstraintError
            If a lifecycle constraint has already been applied to the provided
            method.
        """
        if allow_during and restrict_during or not (allow_during or restrict_during):
            raise ValueError(
                'Must provide exactly one of "allow_during" or "restrict_during".'
            )
        unknown_states = (
            set(allow_during).union(restrict_during).difference(self.lifecycle._state_names)
        )
        if unknown_states:
            raise LifeCycleError(
                f"Attempting to constrain {method} with "
                f"states not in the life cycle: {list(unknown_states)}."
            )
        if restrict_during:
            allow_during = [
                s for s in self.lifecycle._state_names if s not in restrict_during
            ]
        if isinstance(allow_during, tuple):
            allow_during = list(allow_during)

        self._make_constraint(method, allow_during)

    def __repr__(self) -> str:
        return f"LifeCycleManager(state={self.current_state})"

    def __str__(self) -> str:
        return str(self.lifecycle)



"""
===========================
Interface Utility Functions
===========================

The functions defined here are used to support the interactive and command-line
interfaces for ``vivarium``.

"""
import functools
from collections.abc import Callable, Generator, Sequence
from datetime import datetime
from pathlib import Path
from typing import Any

import yaml

from vivarium.exceptions import VivariumError


def run_from_ipython() -> bool:
    """Taken from https://stackoverflow.com/questions/5376837/how-can-i-do-an-if-run-from-ipython-test-in-python"""
    try:
        __IPYTHON__  # type: ignore [name-defined]
        return True
    except NameError:
        return False


def log_progress(
    sequence: Sequence[int],
    every: int | None = None,
    size: int | None = None,
    name: str = "Items",
) -> Generator[int, None, None]:
    """Taken from https://github.com/alexanderkuk/log-progress"""
    from IPython.display import display
    from ipywidgets import HTML, IntProgress, VBox

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)  # every 0.5%
    else:
        assert every is not None, "sequence is iterator, set every"

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = "info"
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    display(box)  # type: ignore [no-untyped-call]

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = "{name}: {index} / ?".format(name=name, index=index)
                else:
                    progress.value = index
                    label.value = "{name}: {index} / {size}".format(
                        name=name, index=index, size=size
                    )
            yield record
    except Exception as e:
        progress.bar_style = "danger"
        raise
    else:
        progress.bar_style = "success"
        progress.value = index
        label.value = "{name}: {index}".format(name=name, index=str(index or "?"))


class InteractiveError(VivariumError):
    """Error raised when the Interactive context is in an inconsistent state."""

    pass


def raise_if_not_setup(system_type: str) -> Callable[..., Any]:
    type_error_map = {
        "run": "Simulation must be setup before it can be run",
        "value": "Value pipeline configuration is not complete until the simulation is setup.",
        "event": "Event configuration is not complete until the simulation is setup.",
        "component": "Component configuration is not complete until the simulation is setup.",
        "population": "No population exists until the simulation is setup.",
    }
    err_msg = type_error_map[system_type]

    def method_wrapper(context_method: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(context_method)
        def wrapped_method(*args: Any, **kwargs: Any) -> Any:
            instance = args[0]
            if not instance._setup:
                raise InteractiveError(err_msg)
            return context_method(*args, **kwargs)

        return wrapped_method

    return method_wrapper


def get_output_model_name_string(
    artifact_path: str | Path | None,
    model_spec_path: str | Path,
) -> str:
    """Find a good string to use as model name in output path creation.

    Parameters
    ----------
    artifact_path
        Path to the artifact file, if exists, else should be None
    model_spec_path
        Path to the model specification file. This must exist.

    Returns
    -------
        A model name string for use in output labeling.
    """
    if artifact_path:
        model_name = Path(artifact_path).stem
    else:
        with open(model_spec_path) as model_spec_file:
            model_spec = yaml.safe_load(model_spec_file)
        try:
            model_name = Path(model_spec["configuration"]["input_data"]["artifact_path"]).stem
        except KeyError:
            model_name = Path(model_spec_path).stem
    return model_name


def get_output_root(
    results_directory: str | Path,
    model_specification_file: str | Path,
    artifact_path: str | Path,
) -> Path:
    """Create a root directory for output files.

    Parameters
    ----------
    results_directory
        Directory to store the results in.
    model_specification_file
        Path to the model specification file.
    artifact_path
        Path to the artifact file.

    Returns
    -------
        The date-stamped output root directory.
    """
    launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    model_name = get_output_model_name_string(artifact_path, model_specification_file)
    output_root = Path(results_directory) / model_name / launch_time
    return output_root



from __future__ import annotations

import itertools
from typing import Callable

import numpy as np
import pandas as pd
import pytest

from tests.framework.results.helpers import BASE_POPULATION, FAMILIARS, HOUSE_CATEGORIES
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.observation import (
    AddingObservation,
    ConcatenatingObservation,
    Observation,
    StratifiedObservation,
    UnstratifiedObservation,
)


@pytest.fixture
def stratified_observation() -> StratifiedObservation:
    return StratifiedObservation(
        name="stratified_observation_name",
        pop_filter="",
        when="whenevs",
        requires_columns=[],
        requires_values=[],
        results_updater=lambda _, __: pd.DataFrame(),
        results_formatter=lambda _, __: pd.DataFrame(),
        aggregator_sources=None,
        aggregator=lambda _: 0.0,
    )


@pytest.fixture
def concatenating_observation() -> ConcatenatingObservation:
    return ConcatenatingObservation(
        name="concatenating_observation_name",
        pop_filter="",
        when="whenevs",
        requires_columns=["some-col", "some-other-col"],
        requires_values=[],
        results_formatter=lambda _, __: pd.DataFrame(),
    )


@pytest.mark.parametrize(
    "observation_type, is_stratified",
    [
        (StratifiedObservation, True),
        (UnstratifiedObservation, False),
        (AddingObservation, True),
        (ConcatenatingObservation, False),
    ],
)
def test_is_stratified(observation_type: type[Observation], is_stratified: bool) -> None:
    assert observation_type.is_stratified() == is_stratified


@pytest.mark.parametrize(
    "stratifications, aggregator_sources, aggregator",
    [
        # Series or single-column dataframe return
        (("familiar",), ["power_level"], len),
        (("familiar",), [], len),
        (("familiar", "house"), ["power_level"], len),
        (("familiar", "house"), [], len),
        ((), ["power_level"], len),
        ((), [], len),
        # Multiple-column dataframe return
        (("familiar",), ["power_level", "tracked"], sum),
        (("familiar", "house"), ["power_level", "tracked"], sum),
        ((), ["power_level", "tracked"], sum),
    ],
)
def test_stratified_observation__aggregate(
    stratifications: tuple[str, ...],
    aggregator_sources: list[str],
    aggregator: Callable[[pd.DataFrame], float | pd.Series[float]],
    stratified_observation: StratifiedObservation,
) -> None:
    """Test that we are aggregating correctly. There are some nuances here:
    - If aggregator_resources is provided, then simply .apply it to the groups passed in.
    - If no aggregator_resources are provided, then we want a full aggregation of the groups.
    - _aggregate can return either a pd.Series or a pd.DataFrame of any number of columns
    """

    filtered_pop = BASE_POPULATION.copy()
    for stratification in stratifications:
        mapped_col = f"{stratification}_mapped_values"
        filtered_pop[mapped_col] = filtered_pop[stratification]
    groups = ResultsContext()._get_groups(
        stratifications=stratifications, filtered_pop=filtered_pop
    )
    aggregates = stratified_observation._aggregate(
        pop_groups=groups,  # type: ignore [arg-type]
        aggregator_sources=aggregator_sources,
        aggregator=aggregator,
    )
    if aggregator == len:
        if stratifications:
            stratification_idx: set[tuple[str, ...] | str] = (
                set(itertools.product(*(FAMILIARS, HOUSE_CATEGORIES)))
                if "house" in stratifications
                else set(FAMILIARS)
            )
            assert set(aggregates.index) == stratification_idx
            check = pd.Series(aggregates.values == len(BASE_POPULATION) / groups.ngroups)
            assert check.all()
        else:
            assert len(aggregates.values) == 1
            assert aggregates.values[0] == len(BASE_POPULATION)
    else:  # sum aggregator
        assert aggregates.shape[1] == 2
        expected = BASE_POPULATION[["power_level", "tracked"]].sum() / groups.ngroups
        if stratifications:
            stratification_idx = (
                set(itertools.product(*(FAMILIARS, HOUSE_CATEGORIES)))
                if "house" in stratifications
                else set(FAMILIARS)
            )
            assert set(aggregates.index) == stratification_idx
            final = aggregates.sum() / groups.ngroups
            assert isinstance(final, pd.Series)
            assert final.equals(expected)
        else:
            assert len(aggregates.values) == 1
            for col in ["power_level", "tracked"]:
                assert aggregates.loc["all", col] == expected[col]


@pytest.mark.parametrize(
    "aggregates",
    [
        pd.Series(data=[1, 2, 3], index=pd.Index(["a", "b", "c"], name="index")),
        pd.DataFrame({"col1": [1, 2], "strat1": [1, 1], "strat2": ["cat", "dog"]}).set_index(
            ["strat1", "strat2"]
        ),
    ],
)
def test_stratified_observation__format(
    aggregates: pd.DataFrame | pd.Series[float], stratified_observation: StratifiedObservation
) -> None:
    new_aggregates = stratified_observation._format(aggregates=aggregates)
    assert isinstance(new_aggregates, pd.DataFrame)
    if isinstance(aggregates, pd.Series):
        assert new_aggregates.equals(aggregates.to_frame("value"))
    else:
        assert new_aggregates.equals(aggregates)


@pytest.mark.parametrize(
    "aggregates",
    [
        pd.DataFrame(
            {VALUE_COLUMN: [1.0, 2.0, 10.0, 20.0]},
            index=pd.Index(["ones"] * 2 + ["tens"] * 2),
        ),
        pd.DataFrame(
            {VALUE_COLUMN: [1.0, 2.0, 10.0, 20.0, "bad", "bad"]},
            index=pd.MultiIndex.from_arrays(
                [
                    ["foo", "bar", "foo", "bar", "foo", "bar"],
                    ["ones", "ones", "tens", "tens", "zeros", "zeros"],
                ],
                names=["nonsense", "type"],
            ),
        ).query('type!="zeros"'),
    ],
)
def test_stratified_observation__expand_index(
    aggregates: pd.DataFrame, stratified_observation: StratifiedObservation
) -> None:
    full_idx_aggregates = stratified_observation._expand_index(aggregates=aggregates)
    # NOTE: pd.MultiIndex is a subclass of pd.Index, i.e. check for this first!
    if isinstance(aggregates.index, pd.MultiIndex):
        # Check that index is cartesian product of the original index levels
        assert full_idx_aggregates.index.equals(
            pd.MultiIndex.from_product(aggregates.index.levels)
        )
        # Check that existing values did not change
        assert (
            full_idx_aggregates.loc[aggregates.index, VALUE_COLUMN]
            == aggregates[VALUE_COLUMN]
        ).all()
        # Check that missingness was filled in with zeros
        assert (full_idx_aggregates.query('type=="zeros"')[VALUE_COLUMN] == 0).all()
    else:
        assert aggregates.equals(full_idx_aggregates)


@pytest.mark.parametrize(
    "stratifications",
    [
        ("familiar",),
        ("familiar", "house"),
        (),
    ],
)
def test_stratified_observation_results_gatherer(
    stratifications: tuple[str, ...], stratified_observation: StratifiedObservation
) -> None:
    ctx = ResultsContext()
    # Append the post-stratified columns
    filtered_population = BASE_POPULATION.copy()
    for stratification in stratifications:
        mapped_col = f"{stratification}_mapped_values"
        filtered_population[mapped_col] = filtered_population[stratification]
    pop_groups = ctx._get_groups(
        stratifications=stratifications, filtered_pop=filtered_population
    )
    df = stratified_observation.observe(pop_groups, stratifications)
    assert set(df.columns) == set(["value"])
    expected_idx_names = (
        list(stratifications) if len(stratifications) > 0 else ["stratification"]
    )
    assert list(df.index.names) == expected_idx_names


@pytest.mark.parametrize(
    "new_observations",
    [
        pd.DataFrame({"value": [1.0, 2.0]}),
        pd.DataFrame(
            {
                "value": [1.0, 2.0],
                "another_value": [3.0, 4.0],
                "yet_another_value": [5.0, 6.0],
            }
        ),
        pd.DataFrame({"another_value": [3.0, 4.0], "yet_another_value": [5.0, 6.0]}),
    ],
)
def test_adding_observation_results_updater(new_observations: pd.DataFrame) -> None:
    existing_results = pd.DataFrame({"value": [0.0, 0.0]})
    obs = AddingObservation(
        name="adding_observation_name",
        pop_filter="",
        when="whenevs",
        requires_columns=[],
        requires_values=[],
        results_formatter=lambda _, __: pd.DataFrame(),
        aggregator_sources=None,
        aggregator=lambda _: 0.0,
    )
    updated_results = obs.results_updater(existing_results, new_observations)
    if "value" in new_observations.columns:
        assert updated_results.equals(new_observations)
    else:
        assert updated_results.equals(pd.concat([existing_results, new_observations], axis=1))


@pytest.mark.parametrize(
    "new_observations, expected_results",
    [
        (
            pd.DataFrame({"value": ["two", "three"]}),
            pd.DataFrame({"value": ["zero", "one", "two", "three"]}),
        ),
        (
            pd.DataFrame(
                {
                    "another_value": ["foo", "bar"],
                    "yet_another_value": ["cat", "dog"],
                }
            ),
            pd.DataFrame(
                {
                    "value": ["zero", "one", np.nan, np.nan],
                    "another_value": [np.nan, np.nan, "foo", "bar"],
                    "yet_another_value": [np.nan, np.nan, "cat", "dog"],
                }
            ),
        ),
    ],
)
def test_concatenating_observation_results_updater(
    new_observations: pd.DataFrame,
    expected_results: pd.DataFrame,
    concatenating_observation: ConcatenatingObservation,
) -> None:
    existing_results = pd.DataFrame({"value": ["zero", "one"]})
    updated_results = concatenating_observation.results_updater(
        existing_results, new_observations
    )
    assert updated_results.equals(expected_results)



"""
=================
Logging Interface
=================

This module provides an interface to the :class:`LoggingManager <vivarium.framework.logging.manager.LoggingManager>`.


"""


from __future__ import annotations

import loguru

from vivarium.framework.logging.manager import LoggingManager
from vivarium.manager import Interface


class LoggingInterface(Interface):
    def __init__(self, manager: LoggingManager) -> None:
        self._manager = manager

    def get_logger(self, component_name: str | None = None) -> loguru.Logger:
        return self._manager.get_logger(component_name)



from pathlib import Path
from typing import Any

import pytest
import pytest_mock
import yaml

from vivarium.framework.configuration import (
    DEFAULT_PLUGINS,
    ConfigurationError,
    _get_default_specification,
    build_model_specification,
    build_simulation_configuration,
    validate_model_specification_file,
)


def test_get_default_specification_user_config(
    mocker: pytest_mock.MockFixture, test_user_config: Path
) -> None:
    expand_user_mock = mocker.patch("vivarium.framework.configuration.Path.expanduser")
    expand_user_mock.return_value = test_user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with("~/vivarium.yaml")

    with test_user_config.open() as f:
        data = {"configuration": yaml.full_load(f)}

    data.update(DEFAULT_PLUGINS)
    data.update({"components": {}})

    assert default_spec.to_dict() == data


def test_get_default_specification_no_user_config(
    mocker: pytest_mock.MockFixture, test_data_dir: Path
) -> None:
    user_config = test_data_dir / "oh_no_nothing_here.yaml"

    expand_user_mock = mocker.patch("vivarium.framework.configuration.Path.expanduser")
    expand_user_mock.return_value = user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with("~/vivarium.yaml")
    data: dict[str, dict[Any, Any]] = {"components": {}, "configuration": {}}
    data.update(DEFAULT_PLUGINS)

    assert default_spec.to_dict() == data


def test_validate_model_specification_failures(
    mocker: pytest_mock.MockFixture, test_data_dir: Path, test_spec: Path
) -> None:
    with pytest.raises(ConfigurationError):
        validate_model_specification_file("made_up_file.yaml")

    with pytest.raises(ConfigurationError):
        model_spec = test_data_dir / "bad_model_specification.txt"
        validate_model_specification_file(model_spec)

    with test_spec.open() as f:
        spec_dict = yaml.full_load(f)
    spec_dict.update({"invalid_key": "some_value"})
    load_mock = mocker.patch("vivarium.framework.configuration.yaml.full_load")
    load_mock.return_value = spec_dict
    with pytest.raises(ConfigurationError):
        validate_model_specification_file(test_spec)


def test_validate_model_specification(test_spec: Path) -> None:
    validate_model_specification_file(test_spec)


def test_build_simulation_configuration(
    mocker: pytest_mock.MockFixture, test_user_config: Path
) -> None:
    expand_user_mock = mocker.patch("vivarium.framework.configuration.Path.expanduser")
    expand_user_mock.return_value = test_user_config

    config = build_simulation_configuration()

    assert expand_user_mock.called_once_with("~/vivarium.yaml")

    with test_user_config.open() as f:
        data = yaml.full_load(f)

    assert config.to_dict() == data


def test_build_model_specification_failure(
    mocker: pytest_mock.MockFixture, test_data_dir: Path, test_spec: Path
) -> None:
    with pytest.raises(ConfigurationError):
        build_model_specification("made_up_file.yaml")

    with pytest.raises(ConfigurationError):
        model_spec = test_data_dir / "bad_model_specification.txt"
        build_model_specification(model_spec)

    with test_spec.open() as f:
        spec_dict = yaml.full_load(f)
    spec_dict.update({"invalid_key": "some_value"})
    load_mock = mocker.patch("vivarium.framework.configuration.yaml.full_load")
    load_mock.return_value = spec_dict
    with pytest.raises(ConfigurationError):
        build_model_specification(str(test_spec))


def test_build_model_specification(
    mocker: pytest_mock.MockFixture, test_spec: Path, test_user_config: Path
) -> None:
    expand_user_mock = mocker.patch("vivarium.framework.configuration.Path.expanduser")
    expand_user_mock.return_value = test_user_config
    loaded_model_spec = build_model_specification(test_spec)

    test_data = DEFAULT_PLUGINS.copy()

    with test_spec.open() as f:
        model_data = yaml.full_load(f)

    test_data.update(model_data)

    with test_user_config.open() as f:
        user_data = yaml.full_load(f)

    test_data["configuration"].update(user_data)

    assert loaded_model_spec.to_dict() == test_data



"""
====================
The Simulation Clock
====================

The components here provide implementations of different kinds of simulation
clocks for use in ``vivarium``.

For more information about time in the simulation, see the associated
:ref:`concept note <time_concept>`.

"""

from __future__ import annotations

import math
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from vivarium.framework.lifecycle import lifecycle_states
from vivarium.types import ClockStepSize, ClockTime

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder
    from vivarium.framework.population.population_view import PopulationView
    from vivarium.framework.event import Event
    from vivarium.framework.population import SimulantData
    from vivarium.framework.values import ValuesManager

from vivarium.framework.values import list_combiner
from vivarium.manager import Interface, Manager


class SimulationClock(Manager):
    """A base clock that includes global clock and a pandas series of clocks for each simulant"""

    @property
    def name(self) -> str:
        return "simulation_clock"

    @property
    def columns_created(self) -> list[str]:
        return ["next_event_time", "step_size"]

    @property
    def columns_required(self) -> list[str]:
        return ["tracked"]

    @property
    def time(self) -> ClockTime:
        """The current simulation time."""
        if self._clock_time is None:
            raise ValueError("No start time provided")
        return self._clock_time

    @property
    def stop_time(self) -> ClockTime:
        """The time at which the simulation will stop."""
        if self._stop_time is None:
            raise ValueError("No stop time provided")
        return self._stop_time

    @property
    def minimum_step_size(self) -> ClockStepSize:
        """The minimum step size."""
        if self._minimum_step_size is None:
            raise ValueError("No minimum step size provided")
        return self._minimum_step_size

    @property
    def standard_step_size(self) -> ClockStepSize:
        """The standard varied step size."""
        if self._standard_step_size is None:
            raise ValueError("No standard step size provided")
        elif self._clock_step_size == 0:
            raise ValueError("Step size cannot be equal to zero")
        return self._standard_step_size

    @property
    def step_size(self) -> ClockStepSize:
        """The size of the next time step."""
        if self._clock_step_size is None:
            raise ValueError("No step size provided")
        elif self._clock_step_size == 0:
            raise ValueError("Step size cannot be equal to zero")
        return self._clock_step_size

    @property
    def event_time(self) -> ClockTime:
        "Convenience method for event time, or clock + step"
        return self.time + self.step_size  # type: ignore [operator]

    @property
    def time_steps_remaining(self) -> int:
        number_steps_remaining = (self.stop_time - self.time) / self.step_size  # type: ignore [operator]
        if not isinstance(number_steps_remaining, (float, int)):
            raise ValueError("Invalid type for number of steps remaining")
        return math.ceil(number_steps_remaining)

    def __init__(self) -> None:
        self._clock_time: ClockTime | None = None
        self._stop_time: ClockTime | None = None
        self._minimum_step_size: ClockStepSize | None = None
        self._standard_step_size: ClockStepSize | None = None
        self._clock_step_size: ClockStepSize | None = None
        self._individual_clocks: PopulationView | None = None
        self._pipeline_name = "simulant_step_size"
        # TODO: Delegate this functionality to "tracked" or similar when appropriate
        self._simulants_to_snooze = pd.Index([])

    def setup(self, builder: "Builder") -> None:
        self._step_size_pipeline = builder.value.register_value_producer(
            self._pipeline_name,
            source=lambda idx: [pd.Series(np.nan, index=idx).astype("timedelta64[ns]")],
            preferred_combiner=list_combiner,
            preferred_post_processor=self.step_size_post_processor,
        )
        self.register_step_modifier = partial(
            builder.value.register_value_modifier,
            self._pipeline_name,
            component=self,
        )
        builder.population.initializes_simulants(self, creates_columns=self.columns_created)
        builder.event.register_listener(lifecycle_states.POST_SETUP, self.on_post_setup)
        self._individual_clocks = builder.population.get_view(
            columns=self.columns_created + self.columns_required
        )

    def on_post_setup(self, event: "Event") -> None:
        if not self._step_size_pipeline.mutators:
            ## No components modify the step size, so we use the default
            ## and remove the population view
            self._individual_clocks = None

    def on_initialize_simulants(self, pop_data: "SimulantData") -> None:
        """Sets the next_event_time and step_size columns for each simulant"""
        if self._individual_clocks:
            clocks_to_initialize = pd.DataFrame(
                {
                    "next_event_time": [self.event_time] * len(pop_data.index),
                    "step_size": [self.step_size] * len(pop_data.index),
                },
                index=pop_data.index,
            )
            self._individual_clocks.update(clocks_to_initialize)

    def simulant_next_event_times(self, index: pd.Index[int]) -> pd.Series[ClockTime]:
        """The next time each simulant will be updated."""
        if not self._individual_clocks:
            return pd.Series(self.event_time, index=index)
        return self._individual_clocks.subview(["next_event_time", "tracked"]).get(index)[
            "next_event_time"
        ]

    def simulant_step_sizes(self, index: pd.Index[int]) -> pd.Series[ClockStepSize]:
        """The step size for each simulant."""
        if not self._individual_clocks:
            return pd.Series(self.step_size, index=index)
        return self._individual_clocks.subview(["step_size", "tracked"]).get(index)[
            "step_size"
        ]

    def step_backward(self) -> None:
        """Rewinds the clock by the current step size."""
        if self._clock_time is None:
            raise ValueError("No start time provided")
        self._clock_time -= self.step_size  # type: ignore [operator]

    def step_forward(self, index: pd.Index[int]) -> None:
        """Advances the clock by the current step size, and updates aligned simulant clocks."""
        self._clock_time += self.step_size  # type: ignore [assignment, operator]
        if self._individual_clocks and not index.empty:
            update_index = self.get_active_simulants(index, self.time)
            clocks_to_update = self._individual_clocks.get(update_index)
            if not clocks_to_update.empty:
                clocks_to_update["step_size"] = self._step_size_pipeline(update_index)
                # Simulants that were flagged to get moved to the end should have a next event time
                # of stop time + 1 minimum timestep
                clocks_to_update.loc[self._simulants_to_snooze, "step_size"] = (
                    self.stop_time + self.minimum_step_size - self.time  # type: ignore [operator]
                )
                # TODO: Delegate this functionality to "tracked" or similar when appropriate
                self._simulants_to_snooze = pd.Index([])
                clocks_to_update["next_event_time"] = (
                    self.time + clocks_to_update["step_size"]
                )
                self._individual_clocks.update(clocks_to_update)
            self._clock_step_size = self.simulant_next_event_times(index).min() - self.time  # type: ignore [operator]

    def get_active_simulants(self, index: pd.Index[int], time: ClockTime) -> pd.Index[int]:
        """Gets population that is aligned with global clock"""
        if index.empty or not self._individual_clocks:
            return index
        next_event_times = self.simulant_next_event_times(index)
        return next_event_times[next_event_times <= time].index

    def move_simulants_to_end(self, index: pd.Index[int]) -> None:
        if self._individual_clocks and not index.empty:
            self._simulants_to_snooze = self._simulants_to_snooze.union(index)

    def step_size_post_processor(self, value: Any, manager: ValuesManager) -> Any:
        """Computes the largest feasible step size for each simulant.

        This is the smallest component-modified step size (rounded down to increments
        of the minimum step size), or the global step size, whichever is larger.
        If no components modify the step size, we default to the global step size.

        Parameters
        ----------
        values
            A list of step sizes

        Returns
        -------
            The largest feasible step size for each simulant
        """

        min_modified = pd.DataFrame(value).min(axis=0).fillna(self.standard_step_size)
        ## Rescale pipeline values to global minimum step size
        discretized_step_sizes = (
            np.floor(min_modified / self.minimum_step_size).replace(0, 1)  # type: ignore [attr-defined, operator]
            * self.minimum_step_size
        )
        ## Make sure we don't get zero
        return discretized_step_sizes


class SimpleClock(SimulationClock):
    """A unitless step-count based simulation clock."""

    CONFIGURATION_DEFAULTS = {
        "time": {
            "start": 0,
            "end": 100,
            "step_size": 1,
            "standard_step_size": None,
        }
    }

    @property
    def name(self) -> str:
        return "simple_clock"

    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        time = builder.configuration.time
        self._clock_time = time.start
        self._stop_time = time.end
        self._minimum_step_size = time.step_size
        self._standard_step_size = (
            time.standard_step_size if time.standard_step_size else self._minimum_step_size
        )
        self._clock_step_size = self._standard_step_size

    def __repr__(self) -> str:
        return "SimpleClock()"


def get_time_stamp(time: dict[str, int]) -> pd.Timestamp:
    return pd.Timestamp(time["year"], time["month"], time["day"])


class DateTimeClock(SimulationClock):
    """A date-time based simulation clock."""

    CONFIGURATION_DEFAULTS = {
        "time": {
            "start": {"year": 2005, "month": 7, "day": 2},
            "end": {
                "year": 2010,
                "month": 7,
                "day": 2,
            },
            "step_size": 1,  # Days
            "standard_step_size": None,  # Days
        }
    }

    @property
    def name(self) -> str:
        return "datetime_clock"

    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        time = builder.configuration.time
        self._clock_time = get_time_stamp(time.start)
        self._stop_time = get_time_stamp(time.end)
        self._minimum_step_size = pd.Timedelta(
            days=time.step_size // 1, hours=(time.step_size % 1) * 24
        )
        self._standard_step_size = (
            pd.Timedelta(
                days=time.standard_step_size // 1, hours=(time.standard_step_size % 1) * 24
            )
            if time.standard_step_size
            else self._minimum_step_size
        )
        self._clock_step_size = self._minimum_step_size

    def __repr__(self) -> str:
        return "DateTimeClock()"


class TimeInterface(Interface):
    def __init__(self, manager: SimulationClock) -> None:
        self._manager = manager

    def clock(self) -> Callable[[], ClockTime]:
        """Gets a callable that returns the current simulation time."""
        return lambda: self._manager.time

    def step_size(self) -> Callable[[], ClockStepSize]:
        """Gets a callable that returns the current simulation step size."""
        return lambda: self._manager.step_size

    def simulant_next_event_times(self) -> Callable[[pd.Index[int]], pd.Series[ClockTime]]:
        """Gets a callable that returns the next event times for simulants."""
        return self._manager.simulant_next_event_times

    def simulant_step_sizes(self) -> Callable[[pd.Index[int]], pd.Series[ClockStepSize]]:
        """Gets a callable that returns the simulant step sizes."""
        return self._manager.simulant_step_sizes

    def move_simulants_to_end(self) -> Callable[[pd.Index[int]], None]:
        """Gets a callable that moves simulants to the end of the simulation"""
        return self._manager.move_simulants_to_end

    def register_step_size_modifier(
        self,
        modifier: Callable[[pd.Index[int]], pd.Series[ClockStepSize]],
        requires_columns: list[str] = [],
        requires_values: list[str] = [],
        requires_streams: list[str] = [],
    ) -> None:
        """Registers a step size modifier.

        Parameters
        ----------
        modifier
            Modifier of the step size pipeline. Modifiers can take an index
            and should return a series of step sizes.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the  modifier is called.
        requires_streams
            A list of the randomness streams that need to be properly sourced
            before the modifier is called."""
        return self._manager.register_step_modifier(
            modifier=modifier,
            requires_columns=requires_columns,
            requires_values=requires_values,
            requires_streams=requires_streams,
        )



"""
==================
Artifact Interface
==================

This module provides an interface to the :class:`ArtifactManager <vivarium.framework.artifact.manager.ArtifactManager>`.


"""

from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any

import pandas as pd

from vivarium.framework.artifact.manager import ArtifactManager
from vivarium.manager import Interface

if TYPE_CHECKING:
    from vivarium.types import ScalarValue


class ArtifactInterface(Interface):
    """The builder interface for accessing a data artifact."""

    def __init__(self, manager: ArtifactManager) -> None:
        self._manager = manager

    def load(self, entity_key: str, **column_filters: int | str | Sequence[int | str]) -> Any:
        """Loads data associated with a formatted entity key.

        The provided entity key must be of the form
        {entity_type}.{measure} or {entity_type}.{entity_name}.{measure}.

        Here entity_type denotes the kind of entity being described. Examples
        include cause, risk, population, and covariates.

        The entity_name is the name of the specific entity. For example,
        if we had entity_type as cause, we might have entity_name as
        diarrheal_diseases or ischemic_heart_disease.

        Finally, measure is the name of the quantity the data describes.
        Examples of measures are incidence, disability_weight, relative_risk,
        and cost.

        Parameters
        ----------
        entity_key
            The key associated with the expected data.
        column_filters
            Filters that subset the data by a categorical column and then
            remove the column from the raw data. They are supplied as keyword
            arguments to the load method in the form "column=value".

        Returns
        -------
            The data associated with the given key filtered down to the requested subset.
        """
        return self._manager.load(entity_key, **column_filters)

    def value_columns(
        self,
    ) -> Callable[[str | pd.DataFrame | dict[str, list[ScalarValue] | list[str]]], list[str]]:
        """Returns a function that returns the value columns for the given input.

        The function can be called with either a string or a pandas DataFrame.
        If a string is provided, it is interpreted as an artifact key, and the
        value columns for the data stored at that key are returned.

        Returns
        -------
            A function that returns the value columns for the given input.
        """
        return self._manager.value_columns()

    def __repr__(self) -> str:
        return "ArtifactManagerInterface()"



from vivarium.framework.artifact.artifact import Artifact, ArtifactException
from vivarium.framework.artifact.hdf import EntityKey
from vivarium.framework.artifact.interface import ArtifactInterface
from vivarium.framework.artifact.manager import (
    ArtifactManager,
    filter_data,
    parse_artifact_path_config,
    validate_filter_term,
)



"""
=============
Lookup Tables
=============

Simulations tend to require a large quantity of data to run.  :mod:`vivarium`
provides the :class:`Lookup Table <vivarium.framework.lookup.table.LookupTable>`
abstraction to ensure that accurate data can be retrieved when it's needed. It's
a callable object that takes in a population index and returns data specific to
the individuals represented by that index. See the
:ref:`lookup concept note <lookup_concept>` for more.

"""

from collections.abc import Mapping, Sequence
from datetime import datetime, timedelta
from typing import TYPE_CHECKING
from typing import SupportsFloat as Numeric

import pandas as pd

from vivarium.framework.lifecycle import lifecycle_states
from vivarium.framework.lookup.table import (
    CategoricalTable,
    InterpolatedTable,
    LookupTable,
    ScalarTable,
)
from vivarium.manager import Manager
from vivarium.types import LookupTableData

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class LookupTableManager(Manager):
    """Manages complex data in the simulation.

    Notes
    -----
    Client code should never access this class directly. Use ``lookup`` on the
    builder during setup to get references to LookupTable objects.

    """

    CONFIGURATION_DEFAULTS = {
        "interpolation": {"order": 0, "validate": True, "extrapolate": True}
    }

    @property
    def name(self) -> str:
        return "lookup_table_manager"

    def setup(self, builder: "Builder") -> None:
        self.tables: dict[int, LookupTable] = {}
        self._pop_view_builder = builder.population.get_view
        self.clock = builder.time.clock()
        self._interpolation_order = builder.configuration.interpolation.order
        self._extrapolate = builder.configuration.interpolation.extrapolate
        self._validate = builder.configuration.interpolation.validate
        self._add_constraint = builder.lifecycle.add_constraint

        builder.lifecycle.add_constraint(
            self.build_table, allow_during=[lifecycle_states.SETUP]
        )

    def build_table(
        self,
        data: LookupTableData,
        key_columns: Sequence[str],
        parameter_columns: Sequence[str],
        value_columns: Sequence[str],
    ) -> LookupTable:
        """Construct a lookup table from input data."""
        table = self._build_table(data, key_columns, parameter_columns, value_columns)
        self._add_constraint(
            table.call,
            restrict_during=[
                lifecycle_states.INITIALIZATION,
                lifecycle_states.SETUP,
                lifecycle_states.POST_SETUP,
            ],
        )
        return table

    def _build_table(
        self,
        data: LookupTableData,
        key_columns: Sequence[str],
        parameter_columns: Sequence[str],
        value_columns: Sequence[str],
    ) -> LookupTable:
        # We don't want to require explicit names for tables, but giving them
        # generic names is useful for introspection.
        table_number = len(self.tables)

        if isinstance(data, Mapping):
            data = pd.DataFrame(data)

        if self._validate:
            validate_build_table_parameters(
                data, key_columns, parameter_columns, value_columns
            )

        # Note datetime catches pandas timestamps
        if isinstance(data, (Numeric, datetime, timedelta, list, tuple)):
            table: LookupTable = ScalarTable(
                table_number=table_number,
                data=data,
                key_columns=key_columns,
                parameter_columns=parameter_columns,
                value_columns=value_columns,
                validate=self._validate,
            )
        elif parameter_columns:
            table = InterpolatedTable(
                table_number=table_number,
                data=data,
                population_view_builder=self._pop_view_builder,
                key_columns=key_columns,
                parameter_columns=parameter_columns,
                value_columns=value_columns,
                interpolation_order=self._interpolation_order,
                clock=self.clock,
                extrapolate=self._extrapolate,
                validate=self._validate,
            )
        else:
            table = CategoricalTable(
                table_number=table_number,
                data=data,
                population_view_builder=self._pop_view_builder,
                key_columns=key_columns,
                value_columns=value_columns,
            )

        self.tables[table_number] = table
        return table

    def __repr__(self) -> str:
        return "LookupTableManager()"


def validate_build_table_parameters(
    data: LookupTableData,
    key_columns: Sequence[str],
    parameter_columns: Sequence[str],
    value_columns: Sequence[str],
) -> None:
    """Makes sure the data format agrees with the provided column layout."""
    if (
        data is None
        or (isinstance(data, pd.DataFrame) and data.empty)
        or (isinstance(data, (list, tuple)) and not data)
    ):
        raise ValueError("Must supply some data")

    acceptable_types = (Numeric, datetime, timedelta, list, tuple, pd.DataFrame)
    if not isinstance(data, acceptable_types):
        raise TypeError(
            f"The only allowable types for data are {acceptable_types}. "
            f"You passed {type(data)}."
        )

    if isinstance(data, (list, tuple)):
        if not value_columns:
            raise ValueError(
                "To invoke scalar view with multiple values, you must supply value_columns"
            )
        if len(value_columns) != len(data):
            raise ValueError(
                "The number of value columns must match the number of values."
                f"You supplied values: {data} and value_columns: {value_columns}"
            )
        if key_columns:
            raise ValueError(
                f"key_columns are not allowed for scalar view: Provided {key_columns}."
            )
        if parameter_columns:
            raise ValueError(
                "parameter_columns are not allowed for scalar view: "
                f"Provided {parameter_columns}."
            )

    if isinstance(data, pd.DataFrame):
        if not key_columns and not parameter_columns:
            raise ValueError(
                "Must supply either key_columns or parameter_columns with a DataFrame."
            )

        bin_edge_columns = []
        for p in parameter_columns:
            bin_edge_columns.extend([f"{p}_start", f"{p}_end"])
        all_parameter_columns = set(parameter_columns) | set(bin_edge_columns)

        if set(key_columns).intersection(all_parameter_columns):
            raise ValueError(
                f"There should be no overlap between key columns: {key_columns} "
                f"and parameter columns: {parameter_columns}."
            )

        lookup_columns = set(key_columns) | all_parameter_columns
        if set(value_columns).intersection(lookup_columns):
            raise ValueError(
                f"There should be no overlap between value columns: {value_columns} "
                f"and key or parameter columns: {lookup_columns}."
            )

        specified_columns = set(key_columns) | set(bin_edge_columns) | set(value_columns)
        if specified_columns.difference(data.columns):
            raise ValueError(
                f"The columns supplied: {specified_columns} must all be "
                f"present in the passed data: {data.columns}"
            )



"""
====================
Life Cycle Interface
====================

This module provides an interface to the :class:`LifeCycleManager <vivarium.framework.lifecycle.manager.LifeCycleManager>`.

"""
from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from vivarium.manager import Interface

if TYPE_CHECKING:
    from vivarium.framework.event import Event
    from vivarium.framework.lifecycle.manager import LifeCycleManager


class LifeCycleInterface(Interface):
    """Interface to the life cycle management system.

    The life cycle management system allows components to constrain
    methods so that they're only available during certain simulation
    life cycle states.

    """

    def __init__(self, manager: LifeCycleManager):
        self._manager = manager

    def add_handlers(self, state: str, handlers: list[Callable[[Event], None]]) -> None:
        """Registers a set of functions to be called during a life cycle state.

        This method does not apply any constraints, rather it is used
        to build up an execution order for introspection.

        Parameters
        ----------
        state
            The name of the state to register the handlers for.
        handlers
            A list of functions that will execute during the state.
        """
        self._manager.add_handlers(state, handlers)

    def add_constraint(
        self,
        method: Callable[..., Any],
        allow_during: tuple[str, ...] | list[str] = (),
        restrict_during: tuple[str, ...] | list[str] = (),
    ) -> None:
        """Constrains a function to be executable only during certain states.

        Parameters
        ----------
        method
            The method to add constraints to.
        allow_during
            An optional list of life cycle states in which the provided
            method is allowed to be called.
        restrict_during
            An optional list of life cycle states in which the provided
            method is restricted from being called.

        Raises
        ------
        ValueError
            If neither ``allow_during`` nor ``restrict_during`` are provided,
            or if both are provided.
        LifeCycleError
            If states provided as arguments are not in the life cycle.
        ConstraintError
            If a life cycle constraint has already been applied to the
            provided method.
        """
        self._manager.add_constraint(method, allow_during, restrict_during)

    def current_state(self) -> Callable[[], str]:
        """Returns a callable that gets the current simulation lifecycle state.

        Returns
        -------
            A callable that returns the current simulation lifecycle state.
        """
        return lambda: self._manager.current_state



"""
===========================
Vivarium Command Line Tools
===========================

``vivarium`` provides the tool :command:`simulate` for running simulations
from the command line.  It provides three subcommands:

.. list-table:: ``simulate`` sub-commands
    :header-rows: 1
    :widths: 30, 40

    *   - Name
        - Description
    *   - | **run**
        - | Runs a single simulation from a model specification file.
    *   - | **test**
        - | Runs an example simulation that comes packaged with ``vivarium``.
          | Useful as an installation test.

For more information, see the :ref:`tutorial <cli_tutorial>` on running
simulations from the command line.

.. click:: vivarium.interface.cli:simulate
   :prog: simulate
   :show-nested:

"""

import os
from pathlib import Path
from time import time

import click
import yaml
from loguru import logger

from vivarium.examples import disease_model
from vivarium.framework.engine import SimulationContext
from vivarium.framework.logging import (
    configure_logging_to_file,
    configure_logging_to_terminal,
)
from vivarium.framework.utilities import handle_exceptions
from vivarium.interface.utilities import get_output_root


@click.group()
def simulate() -> None:
    """A command line utility for running a single simulation.

    You may initiate a new run with the ``run`` sub-command, initiate a test
    run of a provided model specification with the ``test`` subcommand, or
    profile a simulation run with the ``profile`` subcommand.
    """
    pass


@simulate.command()
@click.argument(
    "model_specification", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    "--artifact_path",
    "-i",
    type=click.Path(resolve_path=True),
    help="The path to the artifact data file.",
)
@click.option(
    "--results_directory",
    "-o",
    type=click.Path(resolve_path=True),
    default=Path("~/vivarium_results/").expanduser(),
    help="The directory to write results to. A folder will be created "
    "in this directory with the same name as the configuration file.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="Logs verbosely. Useful for debugging and development.",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Suppresses all logging except for warnings and errors.",
)
@click.option(
    "--pdb",
    "with_debugger",
    is_flag=True,
    help="Drop into python debugger if an error occurs.",
)
def run(
    model_specification: Path,
    artifact_path: Path,
    results_directory: Path,
    verbose: bool,
    quiet: bool,
    with_debugger: bool,
) -> None:
    """Run a simulation from the command line.

    The simulation itself is defined by the given MODEL_SPECIFICATION yaml file.

    Within the results directory, which defaults to ~/vivarium_results if none
    is provided, a subdirectory will be created with the same name as the
    MODEL_SPECIFICATION if one does not exist. Results will be written to a
    further subdirectory named after the start time of the simulation run.
    """
    if verbose and quiet:
        raise click.UsageError("Cannot be both verbose and quiet.")
    verbosity = 1 + int(verbose) - int(quiet)
    configure_logging_to_terminal(verbosity=verbosity, long_format=False)

    start = time()

    results_root = get_output_root(results_directory, model_specification, artifact_path)
    # Update permissions mask (assign to variable to avoid printing previous value)
    _ = os.umask(0o002)
    results_root.mkdir(parents=True, exist_ok=False)
    output_data_root = results_root / "results"
    output_data_root.mkdir(parents=True, exist_ok=False)

    configure_logging_to_file(output_directory=results_root)

    output_data = {"results_directory": str(output_data_root)}
    input_data = {}
    if artifact_path:
        input_data["artifact_path"] = artifact_path
    override_configuration = {"output_data": output_data, "input_data": input_data}

    sim = SimulationContext(
        model_specification=model_specification, configuration=override_configuration
    )
    with open(f"{results_directory}/model_specification.yaml", "w") as f:
        yaml.dump(sim.model_specification.to_dict(), f)

    main = handle_exceptions(sim.run_simulation, logger, with_debugger)
    main()

    # Save out simulation metadata
    metadata = {}
    metadata["random_seed"] = sim.configuration.randomness.random_seed
    metadata["input_draw"] = sim.configuration.input_data.input_draw_number
    metadata["simulation_run_time"] = time() - start
    metadata["artifact_path"] = artifact_path
    with open(results_root / "metadata.yaml", "w") as f:
        yaml.dump(metadata, f, default_flow_style=False)

    logger.info(f"Simulation finished.\nResults written to {str(output_data_root)}")


@simulate.command()
def test() -> None:
    """Run a test simulation using the ``disease_model.yaml`` model specification
    provided in the examples directory.
    """
    configure_logging_to_terminal(verbosity=2, long_format=False)
    model_specification = disease_model.get_model_specification_path()

    sim = SimulationContext(model_specification)

    main = handle_exceptions(sim.run_simulation, logger, with_debugger=False)
    main()

    click.echo()
    click.secho("Installation test successful!", fg="green")



"""
===================
Component Interface
===================

This module provides an interface to the :class:`RandomnessManager <vivarium.framework.randomness.manager.RandomnessManager>`.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

from vivarium.framework.randomness.manager import RandomnessManager
from vivarium.framework.randomness.stream import RandomnessStream
from vivarium.manager import Interface

if TYPE_CHECKING:
    from vivarium import Component


class RandomnessInterface(Interface):
    def __init__(self, manager: RandomnessManager):
        self._manager = manager

    def get_stream(
        self,
        decision_point: str,
        # TODO [MIC-5452]: all calls should have a component
        component: Component | None = None,
        initializes_crn_attributes: bool = False,
    ) -> RandomnessStream:
        """Provides a new source of random numbers for the given decision point.

        ``vivarium`` provides a framework for Common Random Numbers which
        allows for variance reduction when modeling counter-factual scenarios.
        Users interested in causal analysis and comparisons between simulation
        scenarios should be careful to use randomness streams provided by the
        framework wherever randomness is employed.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically, this
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.
        component
            The component that is requesting the randomness stream.
        initializes_crn_attributes
            A flag indicating whether this stream is used to generate key
            initialization information that will be used to identify simulants
            in the Common Random Number framework. These streams cannot be
            copied and should only be used to generate the state table columns
            specified in ``builder.configuration.randomness.key_columns``.

        Returns
        -------
            An entry point into the Common Random Number generation framework.
            The stream provides vectorized access to random numbers and a few
            other utilities.
        """
        return self._manager.get_randomness_stream(
            decision_point, component, initializes_crn_attributes
        )

    def get_seed(self, decision_point: str) -> int:
        """Get a randomly generated seed for use with external randomness tools.

        Parameters
        ----------
        decision_point :
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.

        Returns
        -------
            A seed for a random number generation that is linked to Vivarium's
            common random number framework.
        """
        return self._manager.get_seed(decision_point)

    def register_simulants(self, simulants: pd.DataFrame) -> None:
        """Registers simulants with the Common Random Number Framework.

        Parameters
        ----------
        simulants
            A section of the state table with new simulants and at least the
            columns specified in
            ``builder.configuration.randomness.key_columns``.  This function
            should be called as soon as the key columns are generated.
        """
        self._manager.register_simulants(simulants)



from __future__ import annotations
import numpy as np
import pandas as pd

from vivarium.framework.event import Event
from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData
from vivarium.framework.resource.resource import Resource


class Movement(Component):
    ##############
    # Properties #
    ##############
    CONFIGURATION_DEFAULTS = {
        "field": {
            "width": 1000,
            "height": 1000,
        },
        "movement": {
            "max_speed": 2,
        },
    }

    @property
    def columns_created(self) -> list[str]:
        return ["x", "y", "vx", "vy"]

    @property
    def initialization_requirements(self) -> list[str | Resource]:
        return [self.randomness]


    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration

        self.acceleration = builder.value.register_value_producer(
            "acceleration", source=self.base_acceleration
        )
        self.randomness = builder.randomness.get_stream(self.name)

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def base_acceleration(self, index: pd.Index[int]) -> pd.DataFrame:
        return pd.DataFrame(0.0, columns=["x", "y"], index=index)

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        # Start randomly distributed, with random velocities
        new_population = pd.DataFrame(
            {
                "x": self.config.field.width * self.randomness.get_draw(pop_data.index, "x"),
                "y": self.config.field.height * self.randomness.get_draw(pop_data.index, "y"),
                "vx": ((2 * self.randomness.get_draw(pop_data.index, "vx")) - 1) * self.config.movement.max_speed,
                "vy": ((2 * self.randomness.get_draw(pop_data.index, "vy")) - 1) * self.config.movement.max_speed,
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)

    def on_time_step(self, event: Event) -> None:
        pop = self.population_view.get(event.index)

        acceleration = self.acceleration(event.index)

        # Accelerate and limit velocity
        pop[["vx", "vy"]] += acceleration.rename(columns=lambda c: f"v{c}")
        speed = np.sqrt(np.square(pop.vx) + np.square(pop.vy))
        velocity_scaling_factor = np.where(
            speed > self.config.movement.max_speed,
            self.config.movement.max_speed / speed,
            1.0,
        )
        pop["vx"] *= velocity_scaling_factor
        pop["vy"] *= velocity_scaling_factor

        # Move according to velocity
        pop["x"] += pop.vx
        pop["y"] += pop.vy

        # Loop around boundaries
        pop["x"] = pop.x % self.config.field.width
        pop["y"] = pop.y % self.config.field.height

        self.population_view.update(pop)



import re
from types import MethodType

import numpy as np
import pandas as pd
import pytest
import pytest_mock
from _pytest.logging import LogCaptureFixture
from layered_config_tree.main import LayeredConfigTree
from loguru import logger
from pandas.api.types import CategoricalDtype

from tests.framework.results.helpers import (
    BIN_BINNED_COLUMN,
    BIN_LABELS,
    BIN_SILLY_BIN_EDGES,
    BIN_SOURCE,
    FAMILIARS,
    HARRY_POTTER_CONFIG,
    HOUSE_CATEGORIES,
    NAME,
    NAME_COLUMNS,
    POWER_LEVEL_GROUP_LABELS,
    STUDENT_HOUSES,
    CatBombObserver,
    ExamScoreObserver,
    FullyFilteredHousePointsObserver,
    Hogwarts,
    HogwartsResultsStratifier,
    HousePointsObserver,
    MagicalAttributesObserver,
    NeverObserver,
    NoStratificationsQuidditchWinsObserver,
    QuidditchWinsObserver,
    ValedictorianObserver,
    mock_get_value,
    sorting_hat_serial,
    sorting_hat_vectorized,
    verify_stratification_added,
)
from vivarium.framework.event import Event
from vivarium.framework.lifecycle import lifecycle_states
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.results.observation import AddingObservation, Observation
from vivarium.framework.results.stratification import Stratification, get_mapped_col_name
from vivarium.framework.values import Pipeline
from vivarium.interface.interactive import InteractiveContext
from vivarium.types import ScalarMapper, VectorMapper


@pytest.mark.parametrize(
    "stratifications, default_stratifications, additional_stratifications, excluded_stratifications, expected_stratifications",
    [
        ([], [], [], [], []),
        (
            [],
            ["age", "sex"],
            ["handedness"],
            ["age"],
            ["sex", "handedness"],
        ),
        ([], ["age", "sex"], [], ["age", "sex"], []),
        ([], ["age"], [], ["bogus_exclude_column"], ["age"]),
        (["custom"], ["age", "sex"], [], [], ["custom", "age", "sex"]),
    ],
    ids=[
        "empty_add_empty_exclude",
        "one_add_one_exclude",
        "all_defaults_excluded",
        "bogus_exclude",
        "custom_stratification",
    ],
)
def test__get_stratifications(
    stratifications: list[str],
    default_stratifications: list[str],
    additional_stratifications: list[str],
    excluded_stratifications: list[str],
    expected_stratifications: list[str],
    mocker: pytest_mock.MockFixture,
) -> None:
    ctx = ResultsContext()
    ctx.default_stratifications = default_stratifications
    mgr = ResultsManager()
    mgr.logger = mocker.Mock()
    mocker.patch.object(mgr, "_results_context", ctx)
    # default_stratifications would normally be set via ResultsInterface.set_default_stratifications()
    final_stratifications = mgr._get_stratifications(
        stratifications, additional_stratifications, excluded_stratifications
    )
    assert sorted(final_stratifications) == sorted(expected_stratifications)


#######################################
# Tests for `register_stratification` #
#######################################


@pytest.mark.parametrize(
    "excluded_categories, mapper, is_vectorized",
    [
        ([], sorting_hat_vectorized, True),
        ([], sorting_hat_serial, False),
        (["gryffindor"], sorting_hat_vectorized, True),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper", "excluded_categories"],
)
def test_register_stratification_no_pipelines(
    mocker: pytest_mock.MockFixture,
    excluded_categories: list[str],
    mapper: VectorMapper | ScalarMapper,
    is_vectorized: bool,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    mgr.setup(builder)
    mgr.register_stratification(
        name=NAME,
        categories=HOUSE_CATEGORIES,
        excluded_categories=excluded_categories,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=NAME_COLUMNS,
        requires_values=[],
    )
    assert verify_stratification_added(
        stratifications=mgr._results_context.stratifications,
        name=NAME,
        requires_columns=NAME_COLUMNS,
        requires_values=[],
        categories=HOUSE_CATEGORIES,
        excluded_categories=excluded_categories,
        mapper=mapper,
        is_vectorized=is_vectorized,
    )


@pytest.mark.parametrize(
    "mapper, is_vectorized",
    [
        (sorting_hat_vectorized, True),
        (sorting_hat_serial, False),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_pipelines(
    mocker: pytest_mock.MockFixture, mapper: VectorMapper | ScalarMapper, is_vectorized: bool
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mgr.register_stratification(
        name=NAME,
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=[],
        requires_values=NAME_COLUMNS,
    )

    assert verify_stratification_added(
        stratifications=mgr._results_context.stratifications,
        name=NAME,
        requires_columns=[],
        requires_values=[Pipeline(name) for name in NAME_COLUMNS],
        categories=HOUSE_CATEGORIES,
        excluded_categories=[],
        mapper=mapper,
        is_vectorized=is_vectorized,
    )


@pytest.mark.parametrize(
    "mapper, is_vectorized",
    [
        (sorting_hat_vectorized, True),
        (sorting_hat_serial, False),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_column_and_pipelines(
    mocker: pytest_mock.MockFixture, mapper: VectorMapper | ScalarMapper, is_vectorized: bool
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mocked_column_name = "silly_column"
    mgr.register_stratification(
        name=NAME,
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=[mocked_column_name],
        requires_values=NAME_COLUMNS,
    )

    assert verify_stratification_added(
        stratifications=mgr._results_context.stratifications,
        name=NAME,
        requires_columns=[mocked_column_name],
        requires_values=[Pipeline(name) for name in NAME_COLUMNS],
        categories=HOUSE_CATEGORIES,
        excluded_categories=[],
        mapper=mapper,
        is_vectorized=is_vectorized,
    )


##############################################
# Tests for `register_binned_stratification` #
##############################################


@pytest.mark.parametrize(
    "bins, labels",
    [(BIN_SILLY_BIN_EDGES, BIN_LABELS[1:]), (BIN_SILLY_BIN_EDGES[1:], BIN_LABELS)],
    ids=["too_many_bins", "too_many_labels"],
)
def test_register_binned_stratification_raises_bins_labels_mismatch(
    bins: list[float], labels: list[str]
) -> None:
    mgr = ResultsManager()
    with pytest.raises(
        ValueError,
        match=r"The number of bin edges plus 1 \(\d+\) does not match the number of labels \(\d+\)",
    ):
        mgr.register_binned_stratification(
            target=BIN_SOURCE,
            binned_column=BIN_BINNED_COLUMN,
            bin_edges=bins,
            labels=labels,
            excluded_categories=None,
            target_type="column",
        )


from vivarium.types import VectorMapper


def test_binned_stratification_mapper() -> None:
    mgr = ResultsManager()
    mgr.logger = logger
    mgr.register_binned_stratification(
        target=BIN_SOURCE,
        binned_column=BIN_BINNED_COLUMN,
        bin_edges=BIN_SILLY_BIN_EDGES,
        labels=BIN_LABELS,
        excluded_categories=None,
        target_type="column",
    )
    strat = mgr._results_context.stratifications[BIN_BINNED_COLUMN]
    data = pd.DataFrame([-np.inf] + BIN_SILLY_BIN_EDGES + [np.inf])
    groups = strat.vectorized_mapper(data)
    expected_groups = pd.Series([np.nan] + BIN_LABELS + [np.nan] * 2)
    assert (groups.isna() == expected_groups.isna()).all()
    assert (groups[groups.notna()] == expected_groups[expected_groups.notna()]).all()


@pytest.mark.parametrize(
    "default, additional, excluded, match",
    [
        (["age", "sex"], ["age"], [], ["age"]),
        (["age", "sex"], [], ["eye_color"], ["eye_color"]),
        (["age", "sex"], ["age"], ["eye_color"], ["age", "eye_color"]),
    ],
    ids=[
        "additional_no_operation",
        "exclude_no_operation",
        "additional_and_exclude_no_operation",
    ],
)
def test_add_observation_nop_stratifications(
    default: list[str],
    additional: list[str],
    excluded: list[str],
    match: list[str],
    mocker: pytest_mock.MockFixture,
    caplog: LogCaptureFixture,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    mgr.logger = logger

    mgr._results_context.default_stratifications = default
    mgr.register_observation(
        observation_type=AddingObservation,
        name="name",
        pop_filter='alive == "alive"',
        aggregator_sources=[],
        aggregator=lambda: None,
        requires_columns=[],
        requires_values=[],
        additional_stratifications=additional,
        excluded_stratifications=excluded,
        when=lifecycle_states.COLLECT_METRICS,
        results_formatter=lambda: None,
    )
    for m in match:
        assert m in caplog.text


def test_setting_default_stratifications_at_setup(mocker: pytest_mock.MockFixture) -> None:
    """Test that set default stratifications happens at setup"""
    mgr = ResultsManager()
    builder = mocker.Mock()
    mocker.patch.object(mgr._results_context, "set_default_stratifications")
    mgr._results_context.set_default_stratifications.assert_not_called()  # type: ignore[attr-defined]

    mgr.setup(builder)

    mgr._results_context.set_default_stratifications.assert_called_once_with(  # type: ignore[attr-defined]
        builder.configuration.stratification.default
    )


################################
# Tests for results processing #
################################


def test__raw_results_initialized_as_empty_dict(mocker: pytest_mock.MockFixture) -> None:
    """Test that raw results are initialized as an empty dictionary"""
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    assert mgr._raw_results == {}


def test_stratified__raw_results_initialization() -> None:
    """Test that raw results are being initialized correctly. We expect a dictionary
    of pd.DataFrames. Each key of the dictionary is an observed measure name and
    the corresponding value is a zeroed-out multiindex pd.DataFrame of that observer's
    stratifications.
    """

    components = [
        HousePointsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]

    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    raw_results = sim._results._raw_results
    assert isinstance(raw_results, dict)
    assert set(raw_results) == set(["house_points", "quidditch_wins"])
    for measure in raw_results:
        result = raw_results[measure]
        assert isinstance(result, pd.DataFrame)
        assert (result[VALUE_COLUMN] == 0).all()
    STUDENT_HOUSES_LIST = list(STUDENT_HOUSES)

    # Check that indexes are as expected

    # Multi-stratification index is type MultiIndex where each layer dtype is Category
    expected_house_points_multi_idx = pd.MultiIndex.from_product(
        [POWER_LEVEL_GROUP_LABELS, STUDENT_HOUSES_LIST],
        names=["power_level", "student_house"],
    )
    # HACK: We need to set the levels to be CategoricalDtype but you can't set that
    # directly on the MultiIndex. Convert to df, set type, convert back
    expected_house_points_idx = (
        pd.DataFrame(index=expected_house_points_multi_idx)
        .reset_index()
        .astype(CategoricalDtype)
        .set_index(["power_level", "student_house"])
        .index
    )
    assert raw_results["house_points"].index.equals(expected_house_points_idx)

    # Single-stratification index is type CategoricalIndex
    expected_quidditch_idx = pd.CategoricalIndex(FAMILIARS, name="familiar")
    assert raw_results["quidditch_wins"].index.equals(expected_quidditch_idx)


def test_no_stratifications__raw_results_initialization() -> None:
    """Test that Observers requesting no stratifications result in a
    single-row DataFrame with 'value' of zero and index labeled 'all'
    """
    components = [Hogwarts(), NoStratificationsQuidditchWinsObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert isinstance(raw_results, pd.DataFrame)
    assert raw_results.shape == (1, 1)
    assert raw_results[VALUE_COLUMN].iat[0] == 0
    assert raw_results.index.equals(pd.Index(["all"]))


def test_observers_with_missing_stratifications_fail() -> None:
    """Test that an error is raised if an Observer requests a stratification
    that never actually gets registered.
    """
    components = [QuidditchWinsObserver(), HousePointsObserver(), Hogwarts()]

    expected_log_msg = re.escape(
        "The following stratifications are used by observers but not registered: \n"
        "['familiar', 'power_level_group', 'student_house']"
    )

    with pytest.raises(ValueError, match=expected_log_msg):
        InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)


def test_unused_stratifications_are_logged(caplog: LogCaptureFixture) -> None:
    """Test that we issue a logger.info warning if Stratifications are registered
    but never actually used by an Observer

    The HogwartsResultsStratifier registers "student_house", "familiar", and
    "power_level_group" stratifiers. However, we will only use the QuidditchWinsObserver
    which only uses "familiar" and the MagicalAttributesObserver which only uses
    "power_level_group". We would thus expect only "student_house" to be logged
    as an unused stratification.

    """
    components = [
        Hogwarts(),
        HogwartsResultsStratifier(),
        QuidditchWinsObserver(),
        MagicalAttributesObserver(),
    ]
    InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)

    log_split = caplog.text.split(
        "The following stratifications are registered but not used by any observers: \n"
    )
    # Check that the log message is present and only exists one time
    assert len(log_split) == 2
    # Check that the log message contains the expected Stratifications
    assert "['student_house']" in log_split[1]


def test_gather_results_with_no_observations(mocker: pytest_mock.MockerFixture) -> None:
    """Test that gather_results short-circuits when there are no observations for an event."""

    mgr = ResultsManager()
    mgr.population_view = mocker.Mock()
    mgr._results_context = mocker.Mock()
    mgr._results_context.get_observations.return_value = []  # type: ignore[attr-defined]

    event = Event(
        name=lifecycle_states.COLLECT_METRICS,
        index=pd.Index([0]),
        user_data={},
        time=0,
        step_size=1,
    )

    mgr.gather_results(event)

    mgr._results_context.get_observations.assert_called_once_with(event)  # type: ignore[attr-defined]
    mgr.population_view.subview.assert_not_called()  # type: ignore[attr-defined]
    mgr._results_context.gather_results.assert_not_called()  # type: ignore[attr-defined]


def test_gather_results_with_empty_index(mocker: pytest_mock.MockerFixture) -> None:
    """Test that gather_results short-circuits when an event has an empty index."""

    mgr = ResultsManager()
    mgr.population_view = mocker.Mock()
    mgr._results_context = mocker.Mock()
    mgr._results_context.get_observations.return_value = [mocker.Mock(spec=AddingObservation)]  # type: ignore[attr-defined]

    event = Event(
        name=lifecycle_states.COLLECT_METRICS,
        index=pd.Index([]),
        user_data={},
        time=0,
        step_size=1,
    )

    mgr.gather_results(event)

    mgr._results_context.get_observations.assert_called_once_with(event)  # type: ignore[attr-defined]
    mgr.population_view.subview.assert_not_called()  # type: ignore[attr-defined]
    mgr._results_context.gather_results.assert_not_called()  # type: ignore[attr-defined]


def test_gather_results_with_different_stratifications_and_to_observes() -> None:
    components = [
        Hogwarts(),
        HogwartsResultsStratifier(),
        NoStratificationsQuidditchWinsObserver(),
        NeverObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)

    initial_raw_results = sim._results._raw_results.copy()

    sim.step()
    pd.testing.assert_frame_equal(
        sim._results._raw_results["never"], initial_raw_results["never"]
    )
    assert (
        sim._results._raw_results["no_stratifications_quidditch_wins"][VALUE_COLUMN] > 0
    ).all()


@pytest.fixture(scope="module")
def prepare_population_sim() -> InteractiveContext:
    return InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=[Hogwarts()])


@pytest.mark.parametrize(
    "observation_requirements, stratification_requirements, expected_columns",
    [
        ([([], [])], [], []),
        ([([], [])], [(["power_level"], [])], ["power_level", "strat_0_mapped_values"]),
        ([(["familiar"], [])], [], ["familiar"]),
        (
            [(["familiar", "house_points"], [])],
            [(["power_level"], [])],
            ["familiar", "house_points", "power_level", "strat_0_mapped_values"],
        ),
        (
            [(["familiar"], []), (["house_points"], [])],
            [(["power_level"], [])],
            ["familiar", "house_points", "power_level", "strat_0_mapped_values"],
        ),
        (
            [(["familiar"], ["grade"])],
            [(["power_level"], ["double_power"])],
            ["familiar", "grade", "power_level", "double_power", "strat_0_mapped_values"],
        ),
        (
            [(["familiar"], [])],
            [(["power_level"], []), (["house_points"], [])],
            [
                "familiar",
                "power_level",
                "house_points",
                "strat_0_mapped_values",
                "strat_1_mapped_values",
            ],
        ),
        (
            [(["current_time", "event_step_size", "familiar"], [])],
            [(["event_time"], [])],
            [
                "current_time",
                "event_time",
                "event_step_size",
                "familiar",
                "strat_0_mapped_values",
            ],
        ),
        (
            [(["train"], [])],
            [(["headmaster"], [])],
            ["train", "headmaster", "strat_0_mapped_values"],
        ),
    ],
    ids=[
        "no_observation_requirements_no_stratifications",
        "no_observation_requirements",
        "no_stratifications",
        "multiple_columns_single_observation",
        "multiple_observations_with_columns",
        "columns_and_values",
        "multiple_stratifications",
        "time_data_and_columns",
        "user_data",
    ],
)
def test_prepare_population(
    prepare_population_sim: InteractiveContext,
    observation_requirements: list[tuple[list[str], list[str]]],
    stratification_requirements: list[tuple[list[str], list[str]]],
    expected_columns: list[str],
) -> None:
    mgr = prepare_population_sim._results
    observations: list[Observation] = [
        AddingObservation(
            name=f"test_observation_{i}",
            pop_filter="",
            when=lifecycle_states.COLLECT_METRICS,
            requires_columns=columns,
            requires_values=[prepare_population_sim.get_value(value) for value in values],
            results_formatter=lambda *_: pd.DataFrame(),
            aggregator_sources=[],
            aggregator=lambda *_: pd.Series(),
        )
        for i, (columns, values) in enumerate(observation_requirements)
    ]
    stratifications = [
        Stratification(
            name=f"strat_{i}",
            categories=["a", "b", "c"],
            excluded_categories=[],
            requires_columns=columns,
            requires_values=[prepare_population_sim.get_value(value) for value in values],
            mapper=lambda x: pd.Series("a", index=x.index),
            is_vectorized=True,
        )
        for i, (columns, values) in enumerate(stratification_requirements)
    ]

    event = Event(
        name=lifecycle_states.COLLECT_METRICS,
        index=prepare_population_sim.get_population().index,
        user_data={
            "train": "Hogwarts Express",
            "headmaster": "Albus Dumbledore",
            "country": "Scotland",
        },
        time=prepare_population_sim._clock.time + prepare_population_sim._clock.step_size,  # type: ignore [operator]
        step_size=prepare_population_sim._clock.step_size,
    )

    population = mgr._prepare_population(event, observations, stratifications)

    assert set(population.columns) == set(["tracked"] + expected_columns)
    if "current_time" in expected_columns:
        assert (population["current_time"] == prepare_population_sim._clock.time).all()
    if "event_time" in expected_columns:
        assert (population["event_time"] == event.time).all()
    if "event_step_size" in expected_columns:
        assert (population["event_step_size"] == event.step_size).all()
    if "train" in expected_columns:
        assert (population["train"] == "Hogwarts Express").all()
    for strat in stratifications:
        assert (
            population[get_mapped_col_name(strat.name)]
            == strat.stratify(population[strat._sources])
        ).all()


def test_stratified_observation_results() -> None:
    components = [
        Hogwarts(),
        CatBombObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    assert (sim.get_results()["cat_bomb"]["value"] == 0.0).all()
    sim.step()
    num_familiars = sim.get_population().groupby(["familiar", "student_house"]).apply(len)
    expected = num_familiars.loc["cat"] ** 1.0
    expected.name = "value"
    expected = expected.sort_values().reset_index()
    expected["student_house"] = expected["student_house"].astype(
        CategoricalDtype(categories=STUDENT_HOUSES)
    )
    assert expected.equals(
        sim.get_results()["cat_bomb"].sort_values("value").reset_index(drop=True)
    )
    sim.step()
    num_familiars = sim.get_population().groupby(["familiar", "student_house"]).apply(len)
    expected = num_familiars.loc["cat"] ** 2.0
    expected.name = "value"
    expected = expected.sort_values().reset_index()
    expected["student_house"] = expected["student_house"].astype(
        CategoricalDtype(categories=STUDENT_HOUSES)
    )
    assert expected.equals(
        sim.get_results()["cat_bomb"].sort_values("value").reset_index(drop=True)
    )
    _assert_standard_index(sim.get_results()["cat_bomb"])


def test_unstratified_observation_results() -> None:
    components = [
        Hogwarts(),
        ValedictorianObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    first_valedictorian = sim.get_results()["valedictorian"]
    assert len(first_valedictorian) == 1
    sim.step()
    second_valedictorian = sim.get_results()["valedictorian"]
    assert len(second_valedictorian) == 1
    assert (
        first_valedictorian["student_id"].iat[0] != second_valedictorian["student_id"].iat[0]
    )
    _assert_standard_index(first_valedictorian)
    _assert_standard_index(second_valedictorian)


def test_concatenating_observation_results() -> None:
    components = [
        Hogwarts(),
        ExamScoreObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    results_one_step = sim.get_results()["exam_score"]
    assert (results_one_step["exam_score"] == 10.0).all()
    sim.step()
    expected_two_steps = results_one_step.copy()
    expected_two_steps["exam_score"] = 20.0
    expected_two_steps["event_time"] = results_one_step["event_time"] + pd.Timedelta(
        days=sim.configuration.time.step_size
    )
    assert sim.get_results()["exam_score"].equals(
        pd.concat([results_one_step, expected_two_steps], axis=0).reset_index(drop=True)
    )
    _assert_standard_index(sim.get_results()["exam_score"])


def test_adding_observation_results() -> None:
    """Test that adding observation results are being updated correctly."""

    def _check_house_points(pop: pd.DataFrame, step_number: int) -> None:
        """We know that house points are stratified by 'student_house' and 'power_level_group'.
        and that each wizard of gryffindor and of level 20 or 80 (which correspond
        to 'low' and 'very high' power level groups) gains a point
        """
        assert set(pop["house_points"]) == set([0, 1])
        assert (pop.loc[pop["house_points"] != 0, "student_house"] == "gryffindor").all()
        assert set(pop.loc[pop["house_points"] != 0, "power_level"]) == set([20, 80])
        group_sizes = pd.DataFrame(
            pop.groupby(["power_level", "student_house"]).size().astype("float"),
            columns=[VALUE_COLUMN],
        )
        raw_results = sim._results._raw_results["house_points"]
        # We cannot use `equals` here because raw results have a MultiIndex where
        # each layer is a Category dtype but pop has object dtype for the relevant columns
        assert (  # type: ignore[union-attr]
            raw_results.loc[(["low", "very high"], "gryffindor"), "value"].values
            == (group_sizes.loc[([20, 80], "gryffindor"), "value"] * step_number).values
        ).all()

    def _check_quidditch_wins(pop: pd.DataFrame, step_number: int) -> None:
        """We know that quidditch wins are stratified by 'familiar'
        and that each wizard with a banana slug familiar gains a point
        """
        assert set(pop["quidditch_wins"]) == set([0, 1])
        assert (pop.loc[pop["quidditch_wins"] != 0, "familiar"] == "banana_slug").all()
        group_sizes = pd.DataFrame(
            pop.groupby(["familiar"]).size().astype("float"), columns=[VALUE_COLUMN]
        )
        raw_results = sim._results._raw_results["quidditch_wins"]
        # We cannot use `equals` here because raw results have a MultiIndex where
        # each layer is a Category dtype but pop has object dtype for the relevant columns
        assert (
            raw_results[raw_results[VALUE_COLUMN] != 0].values
            == (group_sizes[group_sizes.index == "banana_slug"] * step_number).values
        ).all()

    components = [
        Hogwarts(),
        HousePointsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    pop = sim.get_population()
    _check_house_points(pop, step_number=1)
    _check_quidditch_wins(pop, step_number=1)

    sim.step()
    pop = sim.get_population()
    _check_house_points(pop, step_number=2)
    _check_quidditch_wins(pop, step_number=2)
    _assert_standard_index(sim.get_results()["house_points"])
    _assert_standard_index(sim.get_results()["quidditch_wins"])


def test_concatenating_observation_updates() -> None:
    """Test that concatenating observation raw results are being updated correctly."""
    components = [
        Hogwarts(),
        ExamScoreObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    results_one_step = sim.get_results()["exam_score"]
    assert (results_one_step["exam_score"] == 10.0).all()
    sim.step()
    expected_two_steps = results_one_step.copy()
    expected_two_steps["exam_score"] = 20.0
    expected_two_steps["event_time"] = results_one_step["event_time"] + pd.Timedelta(
        days=sim.configuration.time.step_size
    )
    assert sim.get_results()["exam_score"].equals(
        pd.concat([results_one_step, expected_two_steps], axis=0).reset_index(drop=True)
    )


def test_update__raw_results_fully_filtered_pop() -> None:
    components = [
        Hogwarts(),
        FullyFilteredHousePointsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    # The FullyFilteredHousePointsObserver filters the population to a bogus
    # power level and so we should not be observing anything
    assert (sim._results._raw_results["house_points"][VALUE_COLUMN] == 0).all()
    sim.step()
    assert (sim._results._raw_results["house_points"][VALUE_COLUMN] == 0).all()


def test_update__raw_results_no_stratifications() -> None:
    components = [Hogwarts(), NoStratificationsQuidditchWinsObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    pop = sim.get_population()
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert raw_results.loc["all"][VALUE_COLUMN] == pop["quidditch_wins"].sum()
    sim.step()
    pop = sim.get_population()
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert raw_results.loc["all"][VALUE_COLUMN] == pop["quidditch_wins"].sum() * 2


def test_update__raw_results_extra_columns() -> None:
    """Test that raw results are updated correctly when the aggregator return
    contains multiple columns (i.e. not just a single 'value' column)
    """
    components = [Hogwarts(), HogwartsResultsStratifier(), MagicalAttributesObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    raw_results = sim._results._raw_results["magical_attributes"]
    assert (raw_results[["spell_power", "potion_power"]].values == [1, 1]).all()
    sim.step()
    raw_results = sim._results._raw_results["magical_attributes"]
    assert (raw_results[["spell_power", "potion_power"]].values == [2, 2]).all()


####################
# Helper functions #
####################


def _assert_standard_index(df: pd.DataFrame) -> None:
    """The results should have any stratifications as columns, not indexes."""
    assert not isinstance(df.index, pd.MultiIndex)
    assert df.index.names == [None]
    assert df.index.dtype == "int64"



"""
==================
Resource Interface
==================

This module provides an interface to the :class:`ResourceManager <vivarium.framework.resource.manager.ResourceManager>`.

"""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from vivarium.framework.resource.manager import ResourceManager
from vivarium.framework.resource.resource import Resource
from vivarium.manager import Interface, Manager

if TYPE_CHECKING:
    from vivarium import Component


class ResourceInterface(Interface):
    """The resource management system.

    A resource in :mod:`vivarium` is something like a state table column
    or a randomness stream. These resources are used to initialize or alter
    the state of the simulation. Many of these resources might depend on each
    other and therefore need to be created or updated in a particular order.
    These dependency chains can be quite long and complex.

    Placing the ordering responsibility on end users makes simulations very
    fragile and difficult to understand. Instead, the resource management
    system allows users to only specify local dependencies. The system then
    uses the local dependency information to construct a full dependency
    graph, validate that there are no cyclic dependencies, and return
    resources and their producers in an order that makes sense.

    """

    def __init__(self, manager: ResourceManager):
        self._manager = manager

    def add_resources(
        self,
        # TODO [MIC-5452]: all resource groups should have a component
        component: Component | Manager | None,
        resources: Iterable[str | Resource],
        dependencies: Iterable[str | Resource],
    ) -> None:
        """Adds managed resources to the resource pool.

        Parameters
        ----------
        component
            The component or manager adding the resources.
        resources
            The resources being added. A string represents a column resource.
        dependencies
            A list of resources that the producer requires. A string represents
            a column resource.

        Raises
        ------
        ResourceError
            If either the resource type is invalid, a component has multiple
            resource producers for the ``column`` resource type, or
            there are multiple producers of the same resource.
        """
        self._manager.add_resources(component, resources, dependencies)

    def get_population_initializers(self) -> list[Any]:
        """Returns a dependency-sorted list of population initializers.

        We exclude all non-initializer dependencies. They were necessary in
        graph construction, but we only need the column producers at population
        creation time.
        """
        return self._manager.get_population_initializers()



from pathlib import Path

from vivarium import InteractiveContext
from vivarium.examples.disease_model.disease import (
    DiseaseModel,
    DiseaseState,
    DiseaseTransition,
    SISDiseaseModel,
)
from vivarium.examples.disease_model.intervention import TreatmentIntervention
from vivarium.examples.disease_model.mortality import Mortality
from vivarium.examples.disease_model.observer import DeathsObserver, YllsObserver
from vivarium.examples.disease_model.population import BasePopulation
from vivarium.examples.disease_model.risk import Risk, RiskEffect


def get_model_specification_path() -> str:
    p = Path(__file__).parent / "disease_model.yaml"
    return str(p)


def get_disease_model_simulation() -> InteractiveContext:
    p = get_model_specification_path()
    return InteractiveContext(p)



import pandas as pd
import pytest

from vivarium import InteractiveContext
from vivarium.framework.values import Pipeline


def test_list_values() -> None:
    sim = InteractiveContext()
    # a 'simulant_step_size' value is created by default upon setup
    assert sim.list_values() == ["simulant_step_size"]
    assert isinstance(sim.get_value("simulant_step_size"), Pipeline)
    with pytest.raises(ValueError, match="No value pipeline 'foo' registered."):
        sim.get_value("foo")
    # ensure that 'foo' did not get added to the list of values
    assert sim.list_values() == ["simulant_step_size"]


def test_run_for_duration() -> None:
    sim = InteractiveContext()
    initial_time = sim._clock.time

    sim.run_for(pd.Timedelta("10 days"))
    assert sim._clock.time == initial_time + pd.Timedelta("10 days")  # type: ignore[operator]

    sim.run_for("5 days")
    assert sim._clock.time == initial_time + pd.Timedelta("15 days")  # type: ignore[operator]



"""
=======
Logging
=======

"""

from vivarium.framework.logging.interface import LoggingInterface
from vivarium.framework.logging.manager import LoggingManager
from vivarium.framework.logging.utilities import (
    configure_logging_to_file,
    configure_logging_to_terminal,
    list_loggers,
)



"""
===================
Component Interface
===================

This module provides an interface to the :class:`ComponentManager <vivarium.framework.components.manager.ComponentManager>`.

"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Union

from vivarium import Component
from vivarium.framework.components.manager import ComponentManager
from vivarium.manager import Interface, Manager

if TYPE_CHECKING:
    _ComponentsType = Sequence[Union[Component, Manager, "_ComponentsType"]]


class ComponentInterface(Interface):
    """The builder interface for the component manager system.

    This class defines component manager methods a ``vivarium`` component can
    access from the builder. It provides methods for querying and adding components
    to the :class:`ComponentManager <vivarium.framework.components.manager.ComponentManager>`.
    """

    def __init__(self, manager: ComponentManager):
        self._manager = manager

    def get_component(self, name: str) -> Component | Manager:
        """Get the component that has ``name`` if presently held by the component
        manager. Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.

        Returns
        -------
            A component that has name ``name``.
        """
        return self._manager.get_component(name)

    def get_components_by_type(
        self, component_type: type[Component | Manager] | Sequence[type[Component | Manager]]
    ) -> list[Component | Manager]:
        """Get all components that are an instance of ``component_type``.

        Parameters
        ----------
        component_type
            A component type to retrieve, compared against internal components
            using isinstance().

        Returns
        -------
            A list of components of type ``component_type``.
        """
        return self._manager.get_components_by_type(component_type)

    def list_components(self) -> dict[str, Component | Manager]:
        """Get a mapping of component names to components held by the manager.

        Returns
        -------
            A dictionary mapping component names to components.
        """
        return self._manager.list_components()
